from .avlTreeReservoir import reservoir
import pytest
from freezegun import freeze_time
from random import random


class TestReservoir:
    @pytest.mark.asyncio
    async def test_accumulate(self):
        """Should be able to accumulate measurements."""
        res = reservoir(3)
        res['pushSome']('test1')
        res['pushSome']('test2')
        assert res['size']() == 2
        assert res['at'](0)['data'] == 'test1'
        assert res['at'](1)['data'] == 'test2'

    @pytest.mark.asyncio
    async def test_randomly_remove(self):
        """Should randomly remove old elements from Reservoir."""
        res = reservoir(3)
        res['pushSome'](5)
        res['pushSome'](4)
        res['pushSome'](3)
        res['pushSome'](2)
        res['pushSome'](1)
        assert res['size']() == 3

    @pytest.mark.asyncio
    async def test_calculate_percentiles(self):
        """Should calculate percentiles when Reservoir has 5 elements."""
        res = reservoir(5)
        data = [5, 1, 3, 2, 4]
        for e in data:
            res['pushSome'](e)

        pers1 = res['getPercentile'](75.13)
        pers2 = res['getPercentile'](75.1)
        pers3 = res['getPercentile'](0.05)
        pers4 = res['getPercentile'](50)
        pers5 = res['getPercentile'](75)

        assert pers1 == 4.0052
        assert pers2 == 4.004
        assert pers3 == 1.002
        assert pers4 == 3
        assert pers5 == 4

    @pytest.mark.asyncio
    async def test_return_percentiles_for_actual_records(self):
        """Should return percentiles for actual records only."""
        with freeze_time() as frozen_datetime:
            res = reservoir(15, 60000)
            for item in [5, 15, 20, 35, 40, 50]:
                res['pushSome'](item)
                frozen_datetime.tick(10.001)
            pers50 = res['getPercentile'](50)
            assert pers50 == 35

    @pytest.mark.asyncio
    async def test_run_x_algorithm(self):
        """Should run X algorithm."""
        with freeze_time() as frozen_datetime:
            res = reservoir(15, 60000)
            for i in range(1000):
                item = random()
                res['pushSome'](item)
                frozen_datetime.tick(1.001)
            assert res['size']() == 15
            max_item = res['max']()
            assert max_item['index'] == 999
            frozen_datetime.tick(60)
            res['getPercentile'](50)
            assert res['size']() == 0

    @pytest.mark.asyncio
    async def test_run_z_algorithm(self):
        """Should run Z algorithm."""
        with freeze_time() as frozen_datetime:
            res = reservoir(10, 60000)
            for i in range(3000):
                item = random()
                res['pushSome'](item)
                frozen_datetime.tick(0.1)
            assert res['size']() == 10



from typing import Dict
from datetime import datetime
import math


class Reservoir:
    """FIFO-like reservoir of a fixed size capable
    calculating running sums, min/max, average, msdev and stddev
    msdev and stddev calculation by Naive algorithm
    (Mean square deviation) msdev = sqrt((∑{i = from 1 to n}(Xi)^2 -(∑{i = from 1 to n}Xi)^2 / N) / N)
    (Standard deviation) stddev = sqrt((∑{i = from 1 to n}(Xi)^2 -(∑{i = from 1 to n}Xi)^2 / N) / N - 1)
    link: https://goo.gl/MAEGP2"""

    def __init__(self, size: int, observation_interval_in_ms: int, object: Dict = None):
        """Inits reservoir

        Args:
            size: Reservoir size.
            observation_interval_in_ms: Reservoir observation interval in ms.
            object: Reservoir options.
        """
        if not object:
            self.array = []
            self.size = size
            self._interval = observation_interval_in_ms / size / 1000
            self._queueEndTime = datetime.now().timestamp()
            self._firstQueueIndex = 0
            self._intermediaryRecord = None
            self.statistics = {
                'count': 0,
                'sum': 0,
                'max': None,
                'min': None,
                'average': 0,
                'sumOfSquares': 0,
                'msdev': 0,
                'stddev': 0
            }
        else:
            self.array = object['array']
            self.size = object['size']
            self._interval = object['_interval']
            self._queueEndTime = object['_queueEndTime']
            self._firstQueueIndex = object['_firstQueueIndex']
            self._intermediaryRecord = object['_intermediaryRecord']
            self.statistics = self.check_statistics_on_restore(object['statistics'])

    def fill_array(self, index: int):
        while len(self.array) <= index:
            self.array.append(None)

    def check_statistics_on_restore(self, statistics):
        if statistics['count'] == 0:
            statistics = {
                'count': 0,
                'sum': 0,
                'max': None,
                'min': None,
                'average': None,
                'sumOfSquares': 0,
                'msdev': None,
                'stddev': None
            }
        elif statistics['count'] < 2:
            statistics['msdev'] = None
            statistics['stddev'] = None
        return statistics

    def push_measurement(self, data: float):
        """Add element to reservoir.

        Args:
            data: Data to add.
        """
        if math.isfinite(data):
            self._update_queue()
            self._update_intermediary_record(data)
            self._update_statistics_on_add(data)

    def get_statistics(self):
        """Return reservoir statistics

        Returns:
            Reservoir statistics
        """
        self._update_queue()
        return self.statistics

    def to_plain_object(self):
        self._update_queue()
        return {
            'array': self.array,
            'size': self.size,
            '_interval': self._interval,
            '_queueEndTime': self._queueEndTime,
            '_firstQueueIndex': self._firstQueueIndex,
            '_intermediaryRecord': self._intermediaryRecord,
            'statistics': self.statistics
        }

    def _update_queue(self):
        intervals_count = self._take_time_intervals_count()
        empty_elements_count = self._take_empty_elements_add_count()
        if empty_elements_count > 0:
            self._add_record(empty_elements_count)
            self._queueEndTime += intervals_count * self._interval

    def _take_empty_elements_add_count(self):
        empty_elements_count = self._take_time_intervals_count()
        if empty_elements_count > self.size:
            empty_elements_count = self.size
        return empty_elements_count

    def _take_time_intervals_count(self) -> float:
        time_now = datetime.now().timestamp()
        time_diff = time_now - self._queueEndTime
        time_intervals_count = math.floor(time_diff / self._interval)
        return time_intervals_count

    def _update_running_statistics_on_remove(self, remove_count: int):
        remove_element_index = self._firstQueueIndex + 1
        for i in range(remove_count):
            if remove_element_index >= self.size:
                remove_element_index = 0

            self.fill_array(remove_element_index)
            self._update_statistics_on_remove(self.array[remove_element_index], remove_element_index)
            self.array[remove_element_index] = {
                'count': 0,
                'sum': 0,
                'max': None,
                'min': None,
                'average': 0,
                'sumOfSquares': 0
            }
            remove_element_index += 1
        remove_element_index -= 1
        if remove_element_index < 0:
            remove_element_index = self.size - 1
        return remove_element_index

    def _update_statistics_on_remove(self, remove_element, remove_element_index):
        if remove_element is not None:
            self.statistics['count'] -= remove_element['count']
            self.statistics['sumOfSquares'] -= remove_element['sumOfSquares']
            self.statistics['sum'] -= remove_element['sum']
            self._update_statistics_min_and_max_on_remove(remove_element, remove_element_index)
            if self.statistics['count'] > 0:
                self.statistics['average'] = self.statistics['sum'] / self.statistics['count']
                if self.statistics['count'] > 1:
                    dif_of_sums = self._calculate_difference_of_sums(
                        self.statistics['sumOfSquares'], self.statistics['sum'], self.statistics['count'])
                    self.statistics['msdev'] = float(math.sqrt(dif_of_sums / self.statistics['count']))
                    self.statistics['stddev'] = float(math.sqrt(dif_of_sums / (self.statistics['count'] - 1)))
                else:
                    self.statistics['stddev'] = None
                    self.statistics['msdev'] = None
            else:
                self.statistics['average'] = None
                self.statistics['stddev'] = None
                self.statistics['msdev'] = None

    def _update_statistics_min_and_max_on_remove(self, remove_element, remove_element_index):
        if remove_element['max'] is not None and remove_element['max'] == self.statistics['max']:
            self.statistics['max'] = self._find_max(remove_element_index)

        if remove_element['min'] is not None and remove_element['min'] == self.statistics['min']:
            self.statistics['min'] = self._find_min(remove_element_index)

    def _update_statistics_on_add(self, el):
        if el is not None:
            self.statistics['count'] += 1
            self.statistics['sum'] += el
            self._update_statistics_min_and_max_on_add(el)
            self.statistics['sumOfSquares'] += math.pow(el, 2)
            if self.statistics['count'] > 0:
                self.statistics['average'] = self.statistics['sum'] / self.statistics['count']
                dif_of_sums = self._calculate_difference_of_sums(
                    self.statistics['sumOfSquares'], self.statistics['sum'], self.statistics['count'])
                if self.statistics['count'] > 1:
                    self.statistics['msdev'] = float(math.sqrt(dif_of_sums / self.statistics['count']))
                    self.statistics['stddev'] = float(math.sqrt(dif_of_sums / (self.statistics['count'] - 1)))
                else:
                    self.statistics['stddev'] = None
                    self.statistics['msdev'] = None

    def _update_statistics_min_and_max_on_add(self, el):
        if self.statistics['max'] is None or self.statistics['max'] < el:
            self.statistics['max'] = el

        if self.statistics['min'] is None or self.statistics['min'] > el:
            self.statistics['min'] = el

    def _add_record(self, empty_elements_count):
        if self._intermediaryRecord is not None:
            self.fill_array(self._firstQueueIndex)
            self.array[self._firstQueueIndex] = self._intermediaryRecord
            self._intermediaryRecord = None
        cur_index_in_array = self._update_running_statistics_on_remove(empty_elements_count)
        self._firstQueueIndex = cur_index_in_array

    def _calculate_difference_of_sums(self, sum1, sum2, count):
        return sum1 - math.pow(sum2, 2) / count

    def _update_intermediary_record(self, el):
        if self._intermediaryRecord is None:
            self._intermediaryRecord = {
                'count': 1,
                'sum': el,
                'max': el,
                'min': el,
                'average': el,
                'sumOfSquares': math.pow(el, 2)
            }
        else:
            if self._intermediaryRecord['max'] < el:
                self._intermediaryRecord['max'] = el
            if self._intermediaryRecord['min'] > el:
                self._intermediaryRecord['min'] = el
            self._intermediaryRecord['count'] += 1
            self._intermediaryRecord['sum'] += el
            self._intermediaryRecord['sumOfSquares'] += math.pow(el, 2)

    def _find_min(self, index):
        min = math.inf
        for i in range(len(self.array)):
            el = self.array[i]
            if el is not None and el['min'] is not None and el['min'] < min and i != index:
                min = el['min']
        if min == math.inf:
            return self._intermediaryRecord['min'] if self._intermediaryRecord is not None else None
        return min

    def _find_max(self, index):
        max = -math.inf
        for i in range(len(self.array)):
            el = self.array[i]
            if el is not None and el['max'] is not None and el['max'] > max and i != index:
                max = el['max']
        if max == -math.inf:
            return self._intermediaryRecord['max'] if self._intermediaryRecord is not None else None
        return max



from .connectionHealthMonitor import ConnectionHealthMonitor
from mock import MagicMock, patch
from .metaApiConnection import MetaApiConnection
from .metatraderAccount import MetatraderAccount
from .terminalState import TerminalState
from freezegun import freeze_time
from typing import List
from datetime import datetime
from asyncio import sleep
import pytest


class MockAccount(MetatraderAccount):

    def __init__(self, id):
        super().__init__(MagicMock(), MagicMock(), MagicMock(), MagicMock())
        self._id = id

    @property
    def id(self):
        return self._id


class MockTerminalState(TerminalState):

    def __init__(self):
        self._connected = True
        self._connected_to_broker = True

    def specification(self, symbol: str):
        return {'quoteSessions': {'MONDAY': [{'from': '08:00:00.000', 'to': '17:00:00.000'}]}}

    @property
    def connected(self) -> bool:
        return self._connected

    @property
    def connected_to_broker(self) -> bool:
        return self._connected_to_broker


class MockConnection(MetaApiConnection):

    def __init__(self):
        self._terminalState = MockTerminalState()
        self._account = MockAccount('id')
        self._synchronized = True
        self._subscribed_symbols = ['EURUSD']

    @property
    def account(self):
        return self._account

    @property
    def subscribed_symbols(self) -> List[str]:
        return self._subscribed_symbols

    @property
    def terminal_state(self):
        return self._terminalState

    @property
    def synchronized(self) -> bool:
        return self._synchronized


start_time = '2020-10-05 10:00:00.000'
broker_times = ['2020-10-05 09:00:00.000', '2020-10-10 10:00:00.000']
connection: MetaApiConnection = None
health_monitor: ConnectionHealthMonitor = None
prices = []


@pytest.fixture(autouse=True)
async def run_around_tests():
    with patch('lib.metaApi.connectionHealthMonitor.asyncio.sleep', new=lambda x: sleep(x / 10)):
        global connection
        connection = MockConnection()
        global health_monitor
        health_monitor = ConnectionHealthMonitor(connection)
        health_monitor._quotesHealthy = True
        global prices
        prices = [{
          'symbol': 'EURUSD',
          'brokerTime': broker_times[0],
        }, {
          'symbol': 'EURUSD',
          'brokerTime': broker_times[1],
        }]
        yield


class TestConnectionHealthMonitor:
    @pytest.mark.asyncio
    async def test_return_100_uptime(self):
        """Should return 100 uptime."""
        await health_monitor.on_symbol_price_updated(1, prices[0])
        await sleep(0.2)
        assert health_monitor.uptime == {'5m': 100, '1h': 100, '1d': 100, '1w': 100}

    @pytest.mark.asyncio
    async def test_return_average_uptime(self):
        """Should return average uptime."""
        with freeze_time(start_time) as frozen_datetime:
            await health_monitor.on_symbol_price_updated(1, prices[0])
            await sleep(0.6)
            frozen_datetime.tick(60)
            await sleep(0.4)
            assert health_monitor.uptime == {'5m': 60, '1h': 60, '1d': 60, '1w': 60}

    @pytest.mark.asyncio
    async def test_check_downtime(self):
        """Should check connection for downtime."""
        await health_monitor.on_symbol_price_updated(1, prices[0])
        await sleep(0.22)
        assert health_monitor.uptime == {'5m': 100, '1h': 100, '1d': 100, '1w': 100}
        connection.terminal_state._connected = False
        await sleep(0.22)
        assert health_monitor.uptime == {'5m': 50, '1h': 50, '1d': 50, '1w': 50}
        connection.terminal_state._connected = True
        connection.terminal_state._connected_to_broker = False
        await sleep(0.42)
        assert health_monitor.uptime == {'5m': 25, '1h': 25, '1d': 25, '1w': 25}
        connection.terminal_state._connected_to_broker = True
        connection._synchronized = False
        await sleep(0.22)
        assert health_monitor.uptime == {'5m': 20, '1h': 20, '1d': 20, '1w': 20}
        connection._synchronized = True
        await sleep(0.61)
        assert health_monitor.uptime == {'5m': 50, '1h': 50, '1d': 50, '1w': 50}

    @pytest.mark.asyncio
    async def test_return_ok(self):
        """Should return ok status."""
        assert health_monitor.health_status == {
            'connected': True,
            'connectedToBroker': True,
            'healthy': True,
            'message': 'Connection to broker is stable. No health issues detected.',
            'quoteStreamingHealthy': True,
            'synchronized': True
        }

    @pytest.mark.asyncio
    async def test_return_one_error_message(self):
        """Should return error status with one message."""
        connection.terminal_state._connected_to_broker = False
        assert health_monitor.health_status == {
            'connected': True,
            'connectedToBroker': False,
            'healthy': False,
            'message': 'Connection is not healthy because connection to broker is not established or lost.',
            'quoteStreamingHealthy': True,
            'synchronized': True
        }

    @pytest.mark.asyncio
    async def test_return_multiple_error_messages(self):
        """Should return error status with multiple messages."""
        connection.terminal_state._connected = False
        connection.terminal_state._connected_to_broker = False
        connection._synchronized = False
        assert health_monitor.health_status == {
            'connected': False,
            'connectedToBroker': False,
            'healthy': False,
            'message': 'Connection is not healthy because connection to API server is not established or lost and ' +
                       'connection to broker is not established or lost ' +
                       'and local terminal state is not synchronized to broker.',
            'quoteStreamingHealthy': True,
            'synchronized': False
        }

    @pytest.mark.asyncio
    async def test_show_as_healthy(self):
        """Should show as healthy if recently updated and in session."""
        await health_monitor.on_symbol_price_updated(1, prices[0])
        await sleep(0.2)
        assert health_monitor.health_status['quoteStreamingHealthy']

    @pytest.mark.asyncio
    async def test_show_as_not_healthy(self):
        """Should show as not healthy if old update and in session."""
        with freeze_time(start_time) as frozen_datetime:
            await health_monitor.on_symbol_price_updated(1, prices[0])
            frozen_datetime.tick(60)
            await sleep(0.2)
            assert not health_monitor.health_status['quoteStreamingHealthy']

    @pytest.mark.asyncio
    async def test_show_as_healthy_if_not_in_session(self):
        """Should show as healthy if not in session."""
        with freeze_time(start_time) as frozen_datetime:
            await health_monitor.on_symbol_price_updated(1, prices[1])
            frozen_datetime.tick(60)
            await sleep(0.2)
            assert health_monitor.health_status['quoteStreamingHealthy']

    @pytest.mark.asyncio
    async def test_show_as_healthy_if_no_symbols(self):
        """Should show as healthy if no symbols."""
        with freeze_time(start_time) as frozen_datetime:
            connection._subscribed_symbols = []
            await health_monitor.on_symbol_price_updated(1, prices[0])
            frozen_datetime.tick(60)
            await sleep(0.2)
            assert health_monitor.health_status['quoteStreamingHealthy']



from .historyStorage import HistoryStorage
from .models import MetatraderDeal, MetatraderOrder
from typing import List
from abc import abstractmethod


class MemoryHistoryStorageModel(HistoryStorage):
    """Abstract class which defines MetaTrader memory history storage interface."""

    @property
    @abstractmethod
    def deals(self) -> List[MetatraderDeal]:
        """Returns all deals stored in history storage.

        Returns:
            All deals stored in history storage.
        """
        pass

    @property
    @abstractmethod
    def history_orders(self) -> List[MetatraderOrder]:
        """Returns all history orders stored in history storage.

        Returns:
            All history orders stored in history storage.
        """
        pass

    @property
    @abstractmethod
    def last_deal_time_by_instance_index(self) -> dict:
        """Returns times of last deals by instance indices.

        Returns:
            Dictionary of last deal times by instance indices."""
        pass

    @property
    @abstractmethod
    def last_history_order_time_by_instance_index(self) -> dict:
        """Returns times of last history orders by instance indices.

        Returns:
            Dictionary of last history orders times by instance indices."""
        pass



import asyncio
from typing import Dict, List
from datetime import datetime


class PacketOrderer:
    """Class which orders the synchronization packets."""

    def __init__(self, out_of_order_listener, ordering_timeout_in_seconds: float):
        """Inits the class.

        Args:
            out_of_order_listener: A function which will receive out of order packet events.
            ordering_timeout_in_seconds: Packet ordering timeout.
        """
        self._outOfOrderListener = out_of_order_listener
        self._orderingTimeoutInSeconds = ordering_timeout_in_seconds
        self._isOutOfOrderEmitted = {}
        self._waitListSizeLimit = 100
        self._outOfOrderInterval = None

    def start(self):
        """Initializes the packet orderer"""
        self._sequenceNumberByInstance = {}
        self._lastSessionStartTimestamp = {}
        self._packetsByInstance = {}

        async def emit_events():
            while True:
                await asyncio.sleep(1)
                self._emit_out_of_order_events()

        if not self._outOfOrderInterval:
            self._outOfOrderInterval = asyncio.create_task(emit_events())

    def stop(self):
        """Deinitializes the packet orderer."""
        self._outOfOrderInterval.cancel()
        self._outOfOrderInterval = None

    def restore_order(self, packet: Dict) -> List[Dict]:
        """Processes the packet and resolves in the order of packet sequence number.

        Args:
            packet: Packet to process.

        """
        instance_id = packet['accountId'] + ':' + str(packet['instanceIndex'] if 'instanceIndex' in packet else 0)
        if 'sequenceNumber' not in packet:
            return [packet]
        if packet['type'] == 'synchronizationStarted' and 'synchronizationId' in packet:
            # synchronization packet sequence just started
            self._isOutOfOrderEmitted[instance_id] = False
            self._sequenceNumberByInstance[instance_id] = packet['sequenceNumber']
            self._lastSessionStartTimestamp[instance_id] = packet['sequenceTimestamp']
            self._packetsByInstance[instance_id] = \
                list(filter(lambda wait_packet: wait_packet['packet']['sequenceTimestamp'] >=
                            packet['sequenceTimestamp'],
                            (self._packetsByInstance[instance_id] if instance_id in
                             self._packetsByInstance else [])))
            return [packet] + self._find_next_packets_from_wait_list(instance_id)
        elif instance_id in self._lastSessionStartTimestamp and \
                packet['sequenceTimestamp'] < self._lastSessionStartTimestamp[instance_id]:
            # filter out previous packets
            return []
        elif instance_id in self._sequenceNumberByInstance and \
                packet['sequenceNumber'] == self._sequenceNumberByInstance[instance_id]:
            # let the duplicate s/n packet to pass through
            return [packet]
        elif instance_id in self._sequenceNumberByInstance and \
                packet['sequenceNumber'] == self._sequenceNumberByInstance[instance_id] + 1:
            # in-order packet was received
            self._sequenceNumberByInstance[instance_id] += 1
            return [packet] + self._find_next_packets_from_wait_list(instance_id)
        else:
            # out-of-order packet was received, add it to the wait list
            self._packetsByInstance[instance_id] = self._packetsByInstance[instance_id] \
                if instance_id in self._packetsByInstance else []
            wait_list = self._packetsByInstance[instance_id]
            wait_list.append({
                'instanceId': instance_id,
                'accountId': packet['accountId'],
                'instanceIndex': packet['instanceIndex'] if 'instanceIndex' in packet else 0,
                'sequenceNumber': packet['sequenceNumber'],
                'packet': packet,
                'receivedAt': datetime.now()
            })
            wait_list.sort(key=lambda i: i['sequenceNumber'])
            while len(wait_list) > self._waitListSizeLimit:
                wait_list.pop(0)
            return []

    def _find_next_packets_from_wait_list(self, instance_id) -> List:
        result = []
        wait_list = self._packetsByInstance[instance_id] if instance_id in self._packetsByInstance else []
        while len(wait_list) and wait_list[0]['sequenceNumber'] in [self._sequenceNumberByInstance[instance_id],
                                                                    self._sequenceNumberByInstance[instance_id] + 1]:
            result.append(wait_list[0]['packet'])
            if wait_list[0]['sequenceNumber'] == self._sequenceNumberByInstance[instance_id] + 1:
                self._sequenceNumberByInstance[instance_id] += 1
            wait_list.pop(0)
        if not len(wait_list) and instance_id in self._packetsByInstance:
            del self._packetsByInstance[instance_id]
        return result

    def _emit_out_of_order_events(self):
        for key, wait_list in self._packetsByInstance.items():
            if len(wait_list) and \
                    (wait_list[0][
                         'receivedAt'].timestamp() + self._orderingTimeoutInSeconds) < datetime.now().timestamp():
                instance_id = wait_list[0]['instanceId']
                if instance_id not in self._isOutOfOrderEmitted or not self._isOutOfOrderEmitted[instance_id]:
                    self._isOutOfOrderEmitted[instance_id] = True
                    # Do not emit onOutOfOrderPacket for packets that come before synchronizationStarted
                    if instance_id in self._sequenceNumberByInstance:
                        asyncio.create_task(self._outOfOrderListener.on_out_of_order_packet(
                            wait_list[0]['accountId'], wait_list[0]['instanceIndex'],
                            self._sequenceNumberByInstance[instance_id] + 1,
                            wait_list[0]['sequenceNumber'], wait_list[0]['packet'], wait_list[0]['receivedAt']))



import os
import asyncio
from metaapi_cloud_sdk import MetaApi
from metaapi_cloud_sdk.clients.metaApi.tradeException import TradeException

# Note: for information on how to use this example code please read https://metaapi.cloud/docs/client/usingCodeExamples

token = os.getenv('TOKEN') or '<put in your token here>'
accountId = os.getenv('ACCOUNT_ID') or '<put in your account id here>'

async def test_meta_api_synchronization():
    api = MetaApi(token)
    try:
        account = await api.metatrader_account_api.get_account(accountId)
        initial_state = account.state
        deployed_states = ['DEPLOYING', 'DEPLOYED']

        if initial_state not in deployed_states:
            #  wait until account is deployed and connected to broker
            print('Deploying account')
            await account.deploy()

        print('Waiting for API server to connect to broker (may take couple of minutes)')
        await account.wait_connected()

        # connect to MetaApi API
        connection = await account.connect()

        # wait until terminal state synchronized to the local state
        print('Waiting for SDK to synchronize to terminal state (may take some time depending on your history size)')
        await connection.wait_synchronized()

        # access local copy of terminal state
        print('Testing terminal state access')
        terminal_state = connection.terminal_state
        print('connected:', terminal_state.connected)
        print('connected to broker:', terminal_state.connected_to_broker)
        print('account information:', terminal_state.account_information)
        print('positions:', terminal_state.positions)
        print('orders:', terminal_state.orders)
        print('specifications:', terminal_state.specifications)
        print('EURUSD specification:', terminal_state.specification('EURUSD'))
        print('EURUSD price:', terminal_state.price('EURUSD'))

        # trade
        print('Submitting pending order')
        try:
            result = await connection.create_limit_buy_order('GBPUSD', 0.07, 1.0, 0.9, 2.0,
                                                             {'comment': 'comm', 'clientId': 'TE_GBPUSD_7hyINWqAlE'})
            print('Trade successful, result code is ' + result['stringCode'])
        except Exception as err:
            print('Trade failed with error:')
            print(api.format_error(err))

        if initial_state not in deployed_states:
            # undeploy account if it was undeployed
            print('Undeploying account')
            await account.undeploy()

    except Exception as err:
        print(api.format_error(err))

asyncio.run(test_meta_api_synchronization())



from .metaApiConnection import MetaApiConnection
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from .models import MetatraderHistoryOrders, MetatraderDeals
from ..clients.metaApi.reconnectListener import ReconnectListener
from ..clients.metaApi.synchronizationListener import SynchronizationListener
from ..clients.timeoutException import TimeoutException
from .metatraderAccount import MetatraderAccount
from datetime import datetime, timedelta
from mock import MagicMock, AsyncMock, patch
from .models import date
from typing import Coroutine
import pytest
import asyncio
from asyncio import sleep


class MockClient(MetaApiWebsocketClient):
    def get_account_information(self, account_id: str) -> asyncio.Future:
        pass

    def get_positions(self, account_id: str) -> asyncio.Future:
        pass

    def get_position(self, account_id: str, position_id: str) -> asyncio.Future:
        pass

    def get_orders(self, account_id: str) -> asyncio.Future:
        pass

    def get_order(self, account_id: str, order_id: str) -> asyncio.Future:
        pass

    def get_history_orders_by_ticket(self, account_id: str, ticket: str) -> MetatraderHistoryOrders:
        pass

    def get_history_orders_by_position(self, account_id: str, position_id: str) -> MetatraderHistoryOrders:
        pass

    def get_history_orders_by_time_range(self, account_id: str, start_time: datetime, end_time: datetime,
                                         offset=0, limit=1000) -> MetatraderHistoryOrders:
        pass

    def get_deals_by_ticket(self, account_id: str, ticket: str) -> MetatraderDeals:
        pass

    def get_deals_by_position(self, account_id: str, position_id: str) -> MetatraderDeals:
        pass

    def get_deals_by_time_range(self, account_id: str, start_time: datetime, end_time: datetime, offset: int = 0,
                                limit: int = 1000) -> MetatraderDeals:
        pass

    def remove_history(self, account_id: str, application: str = None) -> Coroutine:
        pass

    def trade(self, account_id: str, trade) -> asyncio.Future:
        pass

    def reconnect(self, account_id: str):
        pass

    def synchronize(self, account_id: str, instance_index: int, synchronization_id: str,
                    starting_history_order_time: datetime, starting_deal_time: datetime) -> Coroutine:
        pass

    def subscribe(self, account_id: str, instance_index: int = None):
        pass

    def subscribe_to_market_data(self, account_id: str, instance_index: int, symbol: str) -> Coroutine:
        pass

    def unsubscribe_from_market_data(self, account_id: str, instance_index: int, symbol: str) -> Coroutine:
        pass

    def add_synchronization_listener(self, account_id: str, listener):
        pass

    def add_reconnect_listener(self, listener: ReconnectListener):
        pass

    def remove_synchronization_listener(self, account_id: str, listener: SynchronizationListener):
        pass

    def get_symbol_specification(self, account_id: str, symbol: str) -> asyncio.Future:
        pass

    def get_symbol_price(self, account_id: str, symbol: str) -> asyncio.Future:
        pass

    async def wait_synchronized(self, account_id: str, instance_index: int, application_pattern: str,
                                timeout_in_seconds: float):
        pass


class MockAccount(MetatraderAccount):
    @property
    def id(self):
        return 'accountId'

    @property
    def synchronization_mode(self):
        return 'user'


class AutoMockAccount(MetatraderAccount):
    @property
    def id(self):
        return 'accountId'

    @property
    def synchronization_mode(self):
        return 'automatic'


account: MockAccount = None
auto_account: AutoMockAccount = None
client: MockClient = None
api: MetaApiConnection = None


@pytest.fixture(autouse=True)
async def run_around_tests():
    global account
    account = MockAccount(MagicMock(), MagicMock(), MagicMock(), MagicMock())
    global auto_account
    auto_account = AutoMockAccount(MagicMock(), MagicMock(), MagicMock(), MagicMock())
    global client
    client = MockClient('token')
    storage = MagicMock()
    storage.last_history_order_time = AsyncMock(return_value=datetime.now())
    storage.last_deal_time = AsyncMock(return_value=datetime.now())
    global api
    api = MetaApiConnection(client, account, storage, MagicMock())
    yield
    api.health_monitor.stop()
    client._synchronizationThrottler.stop()


class TestMetaApiConnection:
    @pytest.mark.asyncio
    async def test_retrieve_account_information(self):
        """Should retrieve account information."""
        account_information = {
            'broker': 'True ECN Trading Ltd',
            'currency': 'USD',
            'server': 'ICMarketsSC-Demo',
            'balance': 7319.9,
            'equity': 7306.649913200001,
            'margin': 184.1,
            'freeMargin': 7120.22,
            'leverage': 100,
            'marginLevel': 3967.58283542
        }
        client.get_account_information = AsyncMock(return_value=account_information)
        actual = await api.get_account_information()
        assert actual == account_information
        client.get_account_information.assert_called_with('accountId')

    @pytest.mark.asyncio
    async def test_retrieve_positions(self):
        """Should retrieve positions."""
        positions = [{
            'id': '46214692',
            'type': 'POSITION_TYPE_BUY',
            'symbol': 'GBPUSD',
            'magic': 1000,
            'time': '2020-04-15T02:45:06.521Z',
            'updateTime': '2020-04-15T02:45:06.521Z',
            'openPrice': 1.26101,
            'currentPrice': 1.24883,
            'currentTickValue': 1,
            'volume': 0.07,
            'swap': 0,
            'profit': -85.25999999999966,
            'commission': -0.25,
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'stopLoss': 1.17721,
            'unrealizedProfit': -85.25999999999901,
            'realizedProfit': -6.536993168992922e-13
        }]
        client.get_positions = AsyncMock(return_value=positions)
        actual = await api.get_positions()
        assert actual == positions
        client.get_positions.assert_called_with('accountId')

    @pytest.mark.asyncio
    async def test_retrieve_position_by_id(self):
        """Should retrieve position by id."""
        position = {
            'id': '46214692',
            'type': 'POSITION_TYPE_BUY',
            'symbol': 'GBPUSD',
            'magic': 1000,
            'time': '2020-04-15T02:45:06.521Z',
            'updateTime': '2020-04-15T02:45:06.521Z',
            'openPrice': 1.26101,
            'currentPrice': 1.24883,
            'currentTickValue': 1,
            'volume': 0.07,
            'swap': 0,
            'profit': -85.25999999999966,
            'commission': -0.25,
            'clientId': 'TE_GBPUSD_7hyINWqAlE',
            'stopLoss': 1.17721,
            'unrealizedProfit': -85.25999999999901,
            'realizedProfit': -6.536993168992922e-13
        }
        client.get_position = AsyncMock(return_value=position)
        actual = await api.get_position('46214692')
        assert actual == position
        client.get_position.assert_called_with('accountId', '46214692')

    @pytest.mark.asyncio
    async def test_retrieve_orders(self):
        """Should retrieve orders."""
        orders = [{
            'id': '46871284',
            'type': 'ORDER_TYPE_BUY_LIMIT',
            'state': 'ORDER_STATE_PLACED',
            'symbol': 'AUDNZD',
            'magic': 123456,
            'platform': 'mt5',
            'time': '2020-04-20T08:38:58.270Z',
            'openPrice': 1.03,
            'currentPrice': 1.05206,
            'volume': 0.01,
            'currentVolume': 0.01,
            'comment': 'COMMENT2'
        }]
        client.get_orders = AsyncMock(return_value=orders)
        actual = await api.get_orders()
        assert actual == orders
        client.get_orders.assert_called_with('accountId')

    @pytest.mark.asyncio
    async def test_retrieve_order_by_id(self):
        """Should retrieve order by id."""
        order = {
            'id': '46871284',
            'type': 'ORDER_TYPE_BUY_LIMIT',
            'state': 'ORDER_STATE_PLACED',
            'symbol': 'AUDNZD',
            'magic': 123456,
            'platform': 'mt5',
            'time': '2020-04-20T08:38:58.270Z',
            'openPrice': 1.03,
            'currentPrice': 1.05206,
            'volume': 0.01,
            'currentVolume': 0.01,
            'comment': 'COMMENT2'
        }
        client.get_order = AsyncMock(return_value=order)
        actual = await api.get_order('46871284')
        assert actual == order
        client.get_order.assert_called_with('accountId', '46871284')

    @pytest.mark.asyncio
    async def test_retrieve_history_orders_by_ticket(self):
        """Should retrieve history orders by ticket."""
        history_orders = {
            'historyOrders': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'currentPrice': 1.261,
                'currentVolume': 0,
                'doneTime': '2020-04-15T02:45:06.521Z',
                'id': '46214692',
                'magic': 1000,
                'platform': 'mt5',
                'positionId': '46214692',
                'state': 'ORDER_STATE_FILLED',
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.260Z',
                'type': 'ORDER_TYPE_BUY',
                'volume': 0.07
                }],
            'synchronizing': False
        }
        client.get_history_orders_by_ticket = AsyncMock(return_value=history_orders)
        actual = await api.get_history_orders_by_ticket('46214692')
        assert actual == history_orders
        client.get_history_orders_by_ticket.assert_called_with('accountId', '46214692')

    @pytest.mark.asyncio
    async def test_retrieve_history_orders_by_position(self):
        """Should retrieve history orders by position."""
        history_orders = {
            'historyOrders': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'currentPrice': 1.261,
                'currentVolume': 0,
                'doneTime': '2020-04-15T02:45:06.521Z',
                'id': '46214692',
                'magic': 1000,
                'platform': 'mt5',
                'positionId': '46214692',
                'state': 'ORDER_STATE_FILLED',
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.260Z',
                'type': 'ORDER_TYPE_BUY',
                'volume': 0.07
            }],
            'synchronizing': False
        }
        client.get_history_orders_by_position = AsyncMock(return_value=history_orders)
        actual = await api.get_history_orders_by_position('46214692')
        assert actual == history_orders
        client.get_history_orders_by_position.assert_called_with('accountId', '46214692')

    @pytest.mark.asyncio
    async def test_retrieve_history_orders_by_time_range(self):
        """Should retrieve history orders by time range."""
        history_orders = {
            'historyOrders': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'currentPrice': 1.261,
                'currentVolume': 0,
                'doneTime': '2020-04-15T02:45:06.521Z',
                'id': '46214692',
                'magic': 1000,
                'platform': 'mt5',
                'positionId': '46214692',
                'state': 'ORDER_STATE_FILLED',
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.260Z',
                'type': 'ORDER_TYPE_BUY',
                'volume': 0.07
            }],
            'synchronizing': False
        }
        client.get_history_orders_by_time_range = AsyncMock(return_value=history_orders)
        start_time = datetime.now() - timedelta(seconds=1)
        end_time = datetime.now()
        actual = await api.get_history_orders_by_time_range(start_time, end_time, 1, 100)
        assert actual == history_orders
        client.get_history_orders_by_time_range.assert_called_with('accountId', start_time, end_time, 1, 100)

    @pytest.mark.asyncio
    async def test_retrieve_history_deals_by_ticket(self):
        """Should retrieve history deals by ticket."""
        deals = {
            'deals': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'commission': -0.25,
                'entryType': 'DEAL_ENTRY_IN',
                'id': '33230099',
                'magic': 1000,
                'platform': 'mt5',
                'orderId': '46214692',
                'positionId': '46214692',
                'price': 1.26101,
                'profit': 0,
                'swap': 0,
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.521Z',
                'type': 'DEAL_TYPE_BUY',
                'volume': 0.07
            }],
            'synchronizing': False
        }
        client.get_deals_by_ticket = AsyncMock(return_value=deals)
        actual = await api.get_deals_by_ticket('46214692')
        assert actual == deals
        client.get_deals_by_ticket.assert_called_with('accountId', '46214692')

    @pytest.mark.asyncio
    async def test_retrieve_history_deals_by_position(self):
        """Should retrieve history deals by position."""
        deals = {
            'deals': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'commission': -0.25,
                'entryType': 'DEAL_ENTRY_IN',
                'id': '33230099',
                'magic': 1000,
                'platform': 'mt5',
                'orderId': '46214692',
                'positionId': '46214692',
                'price': 1.26101,
                'profit': 0,
                'swap': 0,
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.521Z',
                'type': 'DEAL_TYPE_BUY',
                'volume': 0.07
            }],
            'synchronizing': False
        }
        client.get_deals_by_position = AsyncMock(return_value=deals)
        actual = await api.get_deals_by_position('46214692')
        assert actual == deals
        client.get_deals_by_position.assert_called_with('accountId', '46214692')

    @pytest.mark.asyncio
    async def test_retrieve_history_deals_by_time_range(self):
        """Should retrieve history deals by time range."""
        deals = {
            'deals': [{
                'clientId': 'TE_GBPUSD_7hyINWqAlE',
                'commission': -0.25,
                'entryType': 'DEAL_ENTRY_IN',
                'id': '33230099',
                'magic': 1000,
                'platform': 'mt5',
                'orderId': '46214692',
                'positionId': '46214692',
                'price': 1.26101,
                'profit': 0,
                'swap': 0,
                'symbol': 'GBPUSD',
                'time': '2020-04-15T02:45:06.521Z',
                'type': 'DEAL_TYPE_BUY',
                'volume': 0.07
            }],
            'synchronizing': False
        }
        client.get_deals_by_time_range = AsyncMock(return_value=deals)
        start_time = datetime.now() - timedelta(seconds=1)
        end_time = datetime.now()
        actual = await api.get_deals_by_time_range(start_time, end_time, 1, 100)
        assert actual == deals
        client.get_deals_by_time_range.assert_called_with('accountId', start_time, end_time, 1, 100)

    @pytest.mark.asyncio
    async def test_remove_history(self):
        """Should remove history."""
        client.remove_history = AsyncMock()
        api.history_storage.clear = AsyncMock()
        await api.remove_history('app')
        api.history_storage.clear.assert_called()
        client.remove_history.assert_called_with('accountId', 'app')

    @pytest.mark.asyncio
    async def test_create_market_buy_order(self):
        """Should create market buy order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_market_buy_order('GBPUSD', 0.07, 0.9, 2.0, {'comment': 'comment',
                                                                              'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_BUY', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'stopLoss': 0.9, 'takeProfit': 2.0,
                                                      'comment': 'comment', 'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_create_market_sell_order(self):
        """Should create market sell order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_market_sell_order('GBPUSD', 0.07, 0.9, 2.0, {'comment': 'comment',
                                                                               'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_SELL', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'stopLoss': 0.9, 'takeProfit': 2.0,
                                                      'comment': 'comment', 'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_create_limit_buy_order(self):
        """Should create limit buy order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_limit_buy_order('GBPUSD', 0.07, 1.0, 0.9, 2.0, {'comment': 'comment',
                                                                                  'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_BUY_LIMIT', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'openPrice': 1.0, 'stopLoss': 0.9,
                                                      'takeProfit': 2.0, 'comment': 'comment',
                                                      'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_create_limit_sell_order(self):
        """Should create limit sell order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_limit_sell_order('GBPUSD', 0.07, 1.0, 0.9, 2.0, {'comment': 'comment',
                                                                                   'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_SELL_LIMIT', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'openPrice': 1.0, 'stopLoss': 0.9,
                                                      'takeProfit': 2.0, 'comment': 'comment',
                                                      'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_create_stop_buy_order(self):
        """Should create stop buy order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_stop_buy_order('GBPUSD', 0.07, 1.0, 0.9, 2.0, {'comment': 'comment',
                                                                                 'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_BUY_STOP', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'openPrice': 1.0, 'stopLoss': 0.9,
                                                      'takeProfit': 2.0, 'comment': 'comment',
                                                      'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_create_stop_sell_order(self):
        """Should create stop sell order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_stop_sell_order('GBPUSD', 0.07, 1.0, 0.9, 2.0, {'comment': 'comment',
                                                                                  'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_SELL_STOP', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'openPrice': 1.0, 'stopLoss': 0.9,
                                                      'takeProfit': 2.0, 'comment': 'comment',
                                                      'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_create_stop_limit_buy_order(self):
        """Should create stop limit buy order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_stop_limit_buy_order('GBPUSD', 0.07, 1.5, 1.4, 0.9, 2.0, {
            'comment': 'comment', 'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_BUY_STOP_LIMIT', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'openPrice': 1.5, 'stopLimitPrice': 1.4,
                                                      'stopLoss': 0.9, 'takeProfit': 2.0, 'comment': 'comment',
                                                      'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_create_stop_limit_sell_order(self):
        """Should create stop limit sell order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.create_stop_limit_sell_order('GBPUSD', 0.07, 1.0, 1.1, 2.0, 0.9, {
            'comment': 'comment', 'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_TYPE_SELL_STOP_LIMIT', 'symbol': 'GBPUSD',
                                                      'volume': 0.07, 'openPrice': 1.0, 'stopLimitPrice': 1.1,
                                                      'stopLoss': 2.0, 'takeProfit': 0.9, 'comment': 'comment',
                                                      'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_modify_position(self):
        """Should modify position."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.modify_position('46870472', 2.0, 0.9)
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'POSITION_MODIFY', 'positionId': '46870472',
                                                      'stopLoss': 2.0, 'takeProfit': 0.9})

    @pytest.mark.asyncio
    async def test_close_position_partially(self):
        """Should close position partially."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.close_position_partially('46870472', 0.9)
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'POSITION_PARTIAL', 'positionId': '46870472',
                                                      'volume': 0.9})

    @pytest.mark.asyncio
    async def test_close_position(self):
        """Should close position."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.close_position('46870472')
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'POSITION_CLOSE_ID', 'positionId': '46870472'})

    @pytest.mark.asyncio
    async def test_close_position_by_opposite(self):
        """Should close position by an opposite one."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'positionId': '46870472',
            'closeByPositionId': '46870482'
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.close_by('46870472', '46870482', {'comment': 'comment', 'clientId': 'TE_GBPUSD_7hyINWqAlE'})
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'POSITION_CLOSE_BY', 'positionId': '46870472',
                                                      'closeByPositionId': '46870482', 'comment': 'comment',
                                                      'clientId': 'TE_GBPUSD_7hyINWqAlE'})

    @pytest.mark.asyncio
    async def test_close_positions_by_symbol(self):
        """Should close positions by symbol."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.close_positions_by_symbol('EURUSD')
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'POSITIONS_CLOSE_SYMBOL', 'symbol': 'EURUSD'})

    @pytest.mark.asyncio
    async def test_modify_order(self):
        """Should modify order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.modify_order('46870472', 1.0, 2.0, 0.9)
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_MODIFY', 'orderId': '46870472',
                                                      'openPrice': 1.0, 'stopLoss': 2.0, 'takeProfit': 0.9})

    @pytest.mark.asyncio
    async def test_cancel_order(self):
        """Should cancel order."""
        trade_result = {
            'error': 10009,
            'description': 'TRADE_RETCODE_DONE',
            'orderId': 46870472
        }
        client.trade = AsyncMock(return_value=trade_result)
        actual = await api.cancel_order('46870472')
        assert actual == trade_result
        client.trade.assert_called_with('accountId', {'actionType': 'ORDER_CANCEL', 'orderId': '46870472'})

    @pytest.mark.asyncio
    async def test_reconnect_terminal(self):
        """Should reconnect terminal."""
        client.reconnect = AsyncMock()
        await api.reconnect()
        client.reconnect.assert_called_with('accountId')

    @pytest.mark.asyncio
    async def test_subscribe_to_terminal(self):
        """Should subscribe to terminal."""
        client.subscribe = AsyncMock()

        async def delay_connect():
            await asyncio.sleep(0.1)
            await api.on_connected(1, 1)

        asyncio.create_task(delay_connect())
        await api.subscribe()
        client.subscribe.assert_called_with('accountId')

    @pytest.mark.asyncio
    async def test_retry_subscribe_to_terminal(self):
        """Should retry subscribe to terminal if no response received."""
        with patch('lib.metaApi.metaApiConnection.asyncio.sleep', new=lambda x: sleep(x / 10)):
            response = {'type': 'response', 'accountId': 'accountId', 'requestId': 'requestId'}
            client.subscribe = AsyncMock(side_effect=[TimeoutException('timeout'), response, response])

            async def delay_connect():
                await sleep(0.36)
                await api.on_connected(1, 1)

            asyncio.create_task(delay_connect())
            await api.subscribe()
            client.subscribe.assert_called_with('accountId')
            assert client.subscribe.call_count == 2

    @pytest.mark.asyncio
    async def test_subscribe_after_reconnect(self):
        """Should subscribe after reconnect."""
        with patch('lib.metaApi.metaApiConnection.asyncio.sleep', new=lambda x: sleep(x / 10)):
            client.connect = AsyncMock()
            client.subscribe = AsyncMock()
            asyncio.create_task(api.subscribe())
            await sleep(0.18)
            client.subscribe.assert_called_with('accountId')
            await api.on_connected(1, 1)
            asyncio.create_task(api.on_reconnected())
            await sleep(0.33)
            assert client.subscribe.call_count == 3
            await api.on_connected(1, 1)
            await sleep(0.18)
            assert client.subscribe.call_count == 3

    @pytest.mark.asyncio
    async def test_no_multiple_subscribes(self):
        """Should not send multiple subscribe requests at the same time."""
        client.subscribe = AsyncMock()
        asyncio.create_task(api.subscribe())
        asyncio.create_task(api.subscribe())
        await asyncio.sleep(0.5)
        await api.on_connected(1, 1)
        await asyncio.sleep(0.6)
        client.subscribe.assert_called_with('accountId')
        assert client.subscribe.call_count == 1

    @pytest.mark.asyncio
    async def test_not_retry_subscribe_after_close(self):
        """Should not retry subscribe to terminal if connection is closed."""
        client.subscribe = AsyncMock(side_effect=TimeoutException('timeout'))
        client.unsubscribe = AsyncMock()
        asyncio.create_task(api.subscribe())
        asyncio.create_task(api.close())
        await asyncio.sleep(1)
        client.subscribe.assert_called_with('accountId')
        assert client.subscribe.call_count == 1
        client.unsubscribe.assert_called_with('accountId')

    @pytest.mark.asyncio
    async def test_synchronize_state_with_terminal(self):
        """Should synchronize state with terminal."""
        client.synchronize = AsyncMock()
        with patch('lib.metaApi.metaApiConnection.random_id', return_value='synchronizationId'):
            api = MetaApiConnection(client, account, None, MagicMock())
            await api.history_storage.on_history_order_added(1, {'doneTime': date('2020-01-01T00:00:00.000Z')})
            await api.history_storage.on_deal_added(1, {'time': date('2020-01-02T00:00:00.000Z')})
            await api.synchronize(1)
            client.synchronize.assert_called_with('accountId', 1, 'synchronizationId', date('2020-01-01T00:00:00.000Z'),
                                                  date('2020-01-02T00:00:00.000Z'))

    @pytest.mark.asyncio
    async def test_synchronize_state_with_terminal_from_time(self):
        """Should synchronize state with terminal from specified time."""
        client.synchronize = AsyncMock()
        with patch('lib.metaApi.metaApiConnection.random_id', return_value='synchronizationId'):
            api = MetaApiConnection(client, account, None, MagicMock(), date('2020-10-07T00:00:00.000Z'))
            await api.history_storage.on_history_order_added(1, {'doneTime': date('2020-01-01T00:00:00.000Z')})
            await api.history_storage.on_deal_added(1, {'time': date('2020-01-02T00:00:00.000Z')})
            await api.synchronize(1)
            client.synchronize.assert_called_with('accountId', 1, 'synchronizationId', date('2020-10-07T00:00:00.000Z'),
                                                  date('2020-10-07T00:00:00.000Z'))

    @pytest.mark.asyncio
    async def test_subscribe_to_market_data(self):
        """Should subscribe to market data."""
        client.subscribe_to_market_data = AsyncMock()
        await api.subscribe_to_market_data('EURUSD', 1)
        assert 'EURUSD' in api.subscribed_symbols
        client.subscribe_to_market_data.assert_called_with('accountId', 1, 'EURUSD')

    @pytest.mark.asyncio
    async def test_unsubscribe_from_market_data(self):
        """Should unsubscribe from market data."""
        client.subscribe_to_market_data = AsyncMock()
        client.unsubscribe_from_market_data = AsyncMock()
        await api.subscribe_to_market_data('EURUSD', 1)
        assert 'EURUSD' in api.subscribed_symbols
        await api.unsubscribe_from_market_data('EURUSD', 1)
        assert 'EURUSD' not in api.subscribed_symbols
        client.unsubscribe_from_market_data.assert_called_with('accountId', 1, 'EURUSD')

    @pytest.mark.asyncio
    async def test_retrieve_symbol_specification(self):
        """Should retrieve symbol specification."""
        specification = {
            'symbol': 'AUDNZD',
            'tickSize': 0.00001,
            'minVolume': 0.01,
            'maxVolume': 100,
            'volumeStep': 0.01
        }
        client.get_symbol_specification = AsyncMock(return_value=specification)
        actual = await api.get_symbol_specification('AUDNZD')
        assert actual == specification
        client.get_symbol_specification.assert_called_with('accountId', 'AUDNZD')

    @pytest.mark.asyncio
    async def test_retrieve_symbol_price(self):
        """Should retrieve symbol price."""
        price = {
            'symbol': 'AUDNZD',
            'bid': 1.05297,
            'ask': 1.05309,
            'profitTickValue': 0.59731,
            'lossTickValue': 0.59736
        }
        client.get_symbol_price = AsyncMock(return_value=price)
        actual = await api.get_symbol_price('AUDNZD')
        assert actual == price

        client.get_symbol_price.assert_called_with('accountId', 'AUDNZD')

    @pytest.mark.asyncio
    async def test_save_uptime_stats(self):
        """Should save uptime stats to the server."""
        client.save_uptime = AsyncMock()
        await api.save_uptime({'1h': 100})
        client.save_uptime.assert_called_with('accountId', {'1h': 100})

    @pytest.mark.asyncio
    async def test_initialize(self):
        """Should initialize listeners, terminal state and history storage for accounts with user sync mode."""
        client.add_synchronization_listener = MagicMock()
        api = MetaApiConnection(client, account, MagicMock(), MagicMock())
        assert api.terminal_state
        assert api.history_storage
        client.add_synchronization_listener.assert_any_call('accountId', api)
        client.add_synchronization_listener.assert_any_call('accountId', api.terminal_state)
        client.add_synchronization_listener.assert_any_call('accountId', api.history_storage)

    @pytest.mark.asyncio
    async def test_add_sync_listeners(self):
        """Should add synchronization listeners for account with user synchronization mode."""
        client.add_synchronization_listener = MagicMock()
        api = MetaApiConnection(client, account, MagicMock(), MagicMock())
        listener = {}
        api.add_synchronization_listener(listener)
        client.add_synchronization_listener.assert_called_with('accountId', listener)

    @pytest.mark.asyncio
    async def test_remove_sync_listeners(self):
        """Should remove synchronization listeners for account with user synchronization mode."""
        client.remove_synchronization_listener = MagicMock()
        api = MetaApiConnection(client, account, MagicMock(), MagicMock())
        listener = {}
        api.remove_synchronization_listener(listener)
        client.remove_synchronization_listener.assert_called_with('accountId', listener)

    @pytest.mark.asyncio
    async def test_sync_on_connection(self):
        """Should synchronize on connection."""
        with patch('lib.metaApi.metaApiConnection.random_id', return_value='synchronizationId'):
            client.synchronize = AsyncMock()
            api = MetaApiConnection(client, account, None, MagicMock())
            await api.history_storage.on_history_order_added(1, {'doneTime': date('2020-01-01T00:00:00.000Z')})
            await api.history_storage.on_deal_added(1, {'time': date('2020-01-02T00:00:00.000Z')})
            await api.on_connected(1, 1)
            client.synchronize.assert_called_with('accountId', 1, 'synchronizationId', date('2020-01-01T00:00:00.000Z'),
                                                  date('2020-01-02T00:00:00.000Z'))

    @pytest.mark.asyncio
    async def test_maintain_sync(self):
        """Should maintain synchronization if connection has failed."""
        with patch('lib.metaApi.metaApiConnection.random_id', return_value='synchronizationId'):
            client.synchronize = AsyncMock(side_effect=[Exception('test error'), None])
            api = MetaApiConnection(client, account, None, MagicMock())
            await api.history_storage.on_history_order_added(1, {'doneTime': date('2020-01-01T00:00:00.000Z')})
            await api.history_storage.on_deal_added(1, {'time': date('2020-01-02T00:00:00.000Z')})
            await api.on_connected(1, 1)
            client.synchronize.assert_called_with('accountId', 1, 'synchronizationId', date('2020-01-01T00:00:00.000Z'),
                                                  date('2020-01-02T00:00:00.000Z'))

    @pytest.mark.asyncio
    async def test_restore_market_data_subs_on_sync(self):
        """Should restore market data subscriptions on synchronization."""
        with patch('lib.metaApi.metaApiConnection.random_id', return_value='synchronizationId'):
            client.synchronize = AsyncMock()
            client.subscribe_to_market_data = AsyncMock()
            api = MetaApiConnection(client, account, None, MagicMock())
            await api.history_storage.on_history_order_added(1, {'doneTime': date('2020-01-01T00:00:00.000Z')})
            await api.history_storage.on_deal_added(1, {'time': date('2020-01-02T00:00:00.000Z')})
            await api.subscribe_to_market_data('EURUSD', 1)
            await api.on_connected(1, 1)
            client.synchronize.assert_called_with('accountId', 1, 'synchronizationId', date('2020-01-01T00:00:00.000Z'),
                                                  date('2020-01-02T00:00:00.000Z'))
            client.subscribe_to_market_data.assert_called_with('accountId', 1, 'EURUSD')
            assert client.subscribe_to_market_data.call_count == 2

    @pytest.mark.asyncio
    async def test_unsubscribe_from_events_on_close(self):
        """Should unsubscribe from events on close."""
        client.add_synchronization_listener = MagicMock()
        client.remove_synchronization_listener = MagicMock()
        client.unsubscribe = AsyncMock()
        api = MetaApiConnection(client, account, MagicMock(), MagicMock())
        await api.close()
        client.unsubscribe.assert_any_call('accountId')
        client.remove_synchronization_listener.assert_any_call('accountId', api)
        client.remove_synchronization_listener.assert_any_call('accountId', api.terminal_state)
        client.remove_synchronization_listener.assert_any_call('accountId', api.history_storage)

    @pytest.mark.timeout(60)
    @pytest.mark.asyncio
    async def test_wait_sync_complete_user_mode(self):
        """Should wait until synchronization complete."""
        assert not (await api.is_synchronized(1))
        api._historyStorage.update_disk_storage = AsyncMock()
        try:
            await api.wait_synchronized({'applicationPattern': 'app.*', 'synchronizationId': 'synchronizationId',
                                         'timeoutInSeconds': 1, 'intervalInMilliseconds': 10})
            raise Exception('TimeoutError is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
        await api.on_order_synchronization_finished(1, 'synchronizationId')
        await api.on_deal_synchronization_finished(1, 'synchronizationId')
        promise = api.wait_synchronized({'applicationPattern': 'app.*', 'synchronizationId': 'synchronizationId',
                                         'timeoutInSeconds': 1, 'intervalInMilliseconds': 10})
        start_time = datetime.now()
        await promise
        assert pytest.approx(10, 10) == (datetime.now() - start_time).seconds * 1000
        assert (await api.is_synchronized(1, 'synchronizationId'))

    @pytest.mark.asyncio
    async def test_time_out_waiting_for_sync(self):
        """Should time out waiting for synchronization complete."""
        try:
            await api.wait_synchronized({'applicationPattern': 'app.*', 'synchronizationId': 'synchronizationId',
                                         'timeoutInSeconds': 1, 'intervalInMilliseconds': 10})
            raise Exception('TimeoutError is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
        assert not (await api.is_synchronized(1, 'synchronizationId'))

    @pytest.mark.asyncio
    async def test_subscribe_to_terminal_on_reconnect(self):
        """Should subscribe to terminal on reconnect."""
        api.subscribe = AsyncMock()
        await api.on_reconnected()
        api.subscribe.assert_called_with()

    @pytest.mark.asyncio
    async def test_load_history_storage_from_disk(self):
        """Should load data to history storage from disk."""
        api._historyStorage.initialize = AsyncMock()
        await api.initialize()
        api._historyStorage.initialize.assert_called()



import pytest
from ..httpClient import HttpClient
from .metatraderAccount_client import MetatraderAccountClient
import json
import respx
from httpx import Response


PROVISIONING_API_URL = 'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai'
http_client = HttpClient()
account_client = MetatraderAccountClient(http_client, 'header.payload.sign')


class TestMetatraderAccountClient:
    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_many(self):
        """Should retrieve MetaTrader accounts from API."""
        expected = [{
            '_id': '1eda642a-a9a3-457c-99af-3bc5e8d5c4c9',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud',
            'tags': ['tag1', 'tag2']
        }]
        rsps = respx.get(f'{PROVISIONING_API_URL}/users/current/accounts')\
            .mock(return_value=Response(200, json=expected))
        accounts = await account_client.get_accounts({'provisioningProfileId':
                                                      'f9ce1f12-e720-4b9a-9477-c2d4cb25f076'})
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts' + \
            '?provisioningProfileId=f9ce1f12-e720-4b9a-9477-c2d4cb25f076'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_mt_accounts_with_account_token(self):
        """Should not retrieve MetaTrader accounts from API with account token."""
        account_client = MetatraderAccountClient(http_client, 'token')
        try:
            await account_client.get_accounts({'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076'})
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_accounts method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_one(self):
        """Should retrieve MetaTrader account from API."""
        expected = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud',
            'tags': ['tag1', 'tag2']
        }
        rsps = respx.get(f'{PROVISIONING_API_URL}/users/current/accounts/id') \
            .mock(return_value=Response(200, json=expected))
        accounts = await account_client.get_account('id')
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts/id'
        assert rsps.calls[0].request.method == 'GET'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @respx.mock
    @pytest.mark.asyncio
    async def test_retrieve_by_token(self):
        """Should retrieve MetaTrader account by token from API."""
        expected = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud'
        }
        rsps = respx.get(f'{PROVISIONING_API_URL}/users/current/accounts/accessToken/token') \
            .mock(return_value=Response(200, json=expected))
        account_client = MetatraderAccountClient(http_client, 'token')
        accounts = await account_client.get_account_by_token()
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts/accessToken/token'
        assert rsps.calls[0].request.method == 'GET'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_retrieve_account_by_token_with_api_token(self):
        """Should not retrieve MetaTrader account by token via API with api token."""
        account_client = MetatraderAccountClient(http_client, 'header.payload.sign')
        try:
            await account_client.get_account_by_token()
        except Exception as err:
            assert err.__str__() == 'You can not invoke get_account_by_token method, because you have connected ' + \
                   'with API access token. Please use account access token to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_create(self):
        """Should create MetaTrader account via API."""
        expected = {
            'id': 'id'
        }
        account = {
            'login': '50194988',
            'password': 'Test1234',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'type': 'cloud',
            'tags': ['tag1']
        }
        rsps = respx.post(f'{PROVISIONING_API_URL}/users/current/accounts') \
            .mock(return_value=Response(201, json=expected))
        accounts = await account_client.create_account(account)
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts'
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert accounts == expected

    @pytest.mark.asyncio
    async def test_not_create_mt_account_with_account_token(self):
        """Should not create MetaTrader account via API with account token'."""
        account_client = MetatraderAccountClient(http_client, 'token')
        try:
            await account_client.create_account({})
        except Exception as err:
            assert err.__str__() == 'You can not invoke create_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_deploy(self):
        """Should deploy MetaTrader account via API."""
        rsps = respx.post(f'{PROVISIONING_API_URL}/users/current/accounts/id/deploy') \
            .mock(return_value=Response(204))
        await account_client.deploy_account('id')
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts/id/deploy'
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_deploy_mt_account_with_account_token(self):
        """Should not deploy MetaTrader account via API with account token'."""
        account_client = MetatraderAccountClient(http_client, 'token')
        try:
            await account_client.deploy_account('id')
        except Exception as err:
            assert err.__str__() == 'You can not invoke deploy_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_undeploy(self):
        """Should undeploy MetaTrader account via API."""
        rsps = respx.post(f'{PROVISIONING_API_URL}/users/current/accounts/id/undeploy') \
            .mock(return_value=Response(204))
        await account_client.undeploy_account('id')
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts/id/undeploy'
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_undeploy_mt_account_with_account_token(self):
        """Should not undeploy MetaTrader account via API with account token'."""
        account_client = MetatraderAccountClient(http_client, 'token')
        try:
            await account_client.undeploy_account('id')
        except Exception as err:
            assert err.__str__() == 'You can not invoke undeploy_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_redeploy(self):
        """Should redeploy MetaTrader account via API."""
        rsps = respx.post(f'{PROVISIONING_API_URL}/users/current/accounts/id/redeploy') \
            .mock(return_value=Response(204))
        await account_client.redeploy_account('id')
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts/id/redeploy'
        assert rsps.calls[0].request.method == 'POST'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_redeploy_mt_account_with_account_token(self):
        """Should not redeploy MetaTrader account via API with account token'."""
        account_client = MetatraderAccountClient(http_client, 'token')
        try:
            await account_client.redeploy_account('id')
        except Exception as err:
            assert err.__str__() == 'You can not invoke redeploy_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_delete(self):
        """Should delete MetaTrader account via API."""
        rsps = respx.delete(f'{PROVISIONING_API_URL}/users/current/accounts/id') \
            .mock(return_value=Response(204))
        await account_client.delete_account('id')
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts/id'
        assert rsps.calls[0].request.method == 'DELETE'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'

    @pytest.mark.asyncio
    async def test_not_delete_mt_account_with_account_token(self):
        """Should not delete MetaTrader account via API with account token'."""
        account_client = MetatraderAccountClient(http_client, 'token')
        try:
            await account_client.delete_account('id')
        except Exception as err:
            assert err.__str__() == 'You can not invoke delete_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'

    @respx.mock
    @pytest.mark.asyncio
    async def test_update(self):
        """Should update MetaTrader account via API."""
        update_account = {
              'name': 'new account name',
              'password': 'new_password007',
              'server': 'ICMarketsSC2-Demo',
              'tags': ['tag1']
            }
        rsps = respx.put(f'{PROVISIONING_API_URL}/users/current/accounts/id', json=update_account) \
            .mock(return_value=Response(204))
        await account_client.update_account('id', update_account)
        assert rsps.calls[0].request.url == f'{PROVISIONING_API_URL}/users/current/accounts/id'
        assert rsps.calls[0].request.method == 'PUT'
        assert rsps.calls[0].request.headers['auth-token'] == 'header.payload.sign'
        assert rsps.calls[0].request.content == json.dumps(update_account).encode('utf-8')

    @pytest.mark.asyncio
    async def test_not_update_mt_account_with_account_token(self):
        """Should not update MetaTrader account via API with account token'."""
        account_client = MetatraderAccountClient(http_client, 'token')
        try:
            await account_client.update_account('id', {})
        except Exception as err:
            assert err.__str__() == 'You can not invoke update_account method, because you have connected with ' + \
                                    'account access token. Please use API access token from ' + \
                                    'https://app.metaapi.cloud/token page to invoke this method.'



from ..clients.metaApi.synchronizationListener import SynchronizationListener
from .models import MetatraderOrder, MetatraderDeal
from datetime import datetime
from abc import abstractmethod, ABC


class HistoryStorage(SynchronizationListener, ABC):
    """ Abstract class which defines MetaTrader history storage interface."""

    def __init__(self):
        """Inits the history storage"""
        super().__init__()
        self._orderSynchronizationFinished = {}
        self._dealSynchronizationFinished = {}

    async def initialize(self):
        """Initializes the storage and loads required data from a persistent storage."""
        pass

    @property
    def order_synchronization_finished(self) -> bool:
        """Returns flag indicating whether order history synchronization has finished.

        Returns:
            A flag indicating whether order history synchronization has finished.
        """
        return True in list(self._orderSynchronizationFinished.values())

    @property
    def deal_synchronization_finished(self) -> bool:
        """Returns flag indicating whether deal history synchronization has finished.

        Returns:
            A flag indicating whether order history synchronization has finished.
        """
        return True in list(self._dealSynchronizationFinished.values())

    @abstractmethod
    async def clear(self):
        """Clears the storage and deletes persistent data."""
        pass

    @abstractmethod
    async def last_history_order_time(self, instance_index: int = None) -> datetime:
        """Returns the time of the last history order record stored in the history storage.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
            The time of the last history order record stored in the history storage.
        """
        pass

    @abstractmethod
    async def last_deal_time(self, instance_index: int = None) -> datetime:
        """Returns the time of the last history deal record stored in the history storage.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
            The time of the last history deal record stored in the history storage.
        """
        pass

    @abstractmethod
    async def on_history_order_added(self, instance_index: int, history_order: MetatraderOrder):
        """Invoked when a new MetaTrader history order is added.

        Args:
            instance_index: Index of an account instance connected.
            history_order: New MetaTrader history order.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    @abstractmethod
    async def on_deal_added(self, instance_index: int, deal: MetatraderDeal):
        """Invoked when a new MetaTrader history deal is added.

        Args:
            instance_index: Index of an account instance connected.
            deal: New MetaTrader history deal.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_deal_synchronization_finished(self, instance_index: int, synchronization_id: str):
        """Invoked when a synchronization of history deals on a MetaTrader account have finished.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Synchronization request id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        self._dealSynchronizationFinished[str(instance_index)] = True

    async def on_order_synchronization_finished(self, instance_index: int, synchronization_id: str):
        """Invoked when a synchronization of history orders on a MetaTrader account have finished.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Synchronization request id.

        Returns:
             A coroutine which resolves when the asynchronous event is processed.
        """
        self._orderSynchronizationFinished[str(instance_index)] = True

    async def on_connected(self, instance_index: int, replicas: int):
        """Invoked when connection to MetaTrader terminal established.

        Args:
            instance_index: Index of an account instance connected.
            replicas: Number of account replicas launched.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        self._orderSynchronizationFinished[str(instance_index)] = False
        self._dealSynchronizationFinished[str(instance_index)] = False



from ..clients.metaApi.metatraderDemoAccount_client import MetatraderDemoAccountClient, NewMT4DemoAccount, \
    NewMT5DemoAccount
from .metatraderDemoAccount import MetatraderDemoAccount


class MetatraderDemoAccountApi:
    """Exposes MetaTrader demo account API logic to the consumers."""

    def __init__(self, metatrader_demo_account_client: MetatraderDemoAccountClient):
        """Inits a MetaTrader demo account API instance.

        Args:
            metatrader_demo_account_client: MetaTrader demo account REST API client.
        """
        self._metatraderDemoAccountClient = metatrader_demo_account_client

    async def create_mt4_demo_account(self, profile_id: str, account: NewMT4DemoAccount):
        """Creates new MetaTrader 4 demo account.

        Args:
            profile_id: Id of the provisioning profile that will be used as the basis for creating this account.
            account: Demo account to create.
        """
        demo_account = await self._metatraderDemoAccountClient.create_mt4_demo_account(profile_id, account)
        return MetatraderDemoAccount(demo_account)

    async def create_mt5_demo_account(self, profile_id: str, account: NewMT5DemoAccount):
        """Creates new MetaTrader 5 demo account.

        Args:
            profile_id: Id of the provisioning profile that will be used as the basis for creating this account.
            account: Demo account to create.
        """
        demo_account = await self._metatraderDemoAccountClient.create_mt5_demo_account(profile_id, account)
        return MetatraderDemoAccount(demo_account)



from typing import Dict
from collections import deque
from datetime import datetime
import asyncio


class SynchronizationThrottler:
    """Synchronization throttler used to limit the amount of concurrent synchronizations to prevent application
    from being overloaded due to excessive number of synchronisation responses being sent."""

    def __init__(self, client, max_concurrent_synchronizations: int = 10):
        """Inits the synchronization throttler.

        Args:
            max_concurrent_synchronizations: Limit of concurrent synchronizations.
        """
        self._maxConcurrentSynchronizations = max_concurrent_synchronizations
        self._client = client
        self._synchronizationIds = {}
        self._accountsBySynchronizationIds = {}
        self._synchronizationQueue = deque([])
        self._removeOldSyncIdsInterval = None
        self._processQueueInterval = None

    def start(self):
        """Initializes the synchronization throttler."""
        async def remove_old_sync_ids_interval():
            while True:
                await self._remove_old_sync_ids_job()
                await asyncio.sleep(1)

        async def process_queue_interval():
            while True:
                await self._process_queue_job()
                await asyncio.sleep(1)

        if not self._removeOldSyncIdsInterval:
            self._removeOldSyncIdsInterval = asyncio.create_task(remove_old_sync_ids_interval())
            self._processQueueInterval = asyncio.create_task(process_queue_interval())

    def stop(self):
        """Deinitializes the throttler."""
        self._removeOldSyncIdsInterval.cancel()
        self._removeOldSyncIdsInterval = None
        self._processQueueInterval.cancel()
        self._processQueueInterval = None

    async def _remove_old_sync_ids_job(self):
        now = datetime.now().timestamp()
        for key in list(self._synchronizationIds.keys()):
            if (now - self._synchronizationIds[key]) > 10:
                del self._synchronizationIds[key]
                self._advance_queue()
        while len(self._synchronizationQueue) and \
                (datetime.now().timestamp() - self._synchronizationQueue[0]['queueTime']) > 300:
            self._remove_from_queue(self._synchronizationQueue[0]['synchronizationId'])
        await asyncio.sleep(1)

    def update_synchronization_id(self, synchronization_id: str):
        """Fills a synchronization slot with synchronization id.

        Args:
            synchronization_id: Synchronization id.
        """
        if synchronization_id in self._accountsBySynchronizationIds:
            self._synchronizationIds[synchronization_id] = datetime.now().timestamp()

    @property
    def is_synchronization_available(self) -> bool:
        """Whether there are free slots for synchronization requests."""
        synchronizing_accounts = []
        for key in self._synchronizationIds:
            account_data = self._accountsBySynchronizationIds[key] if key in \
                                                                      self._accountsBySynchronizationIds else None
            if account_data and (account_data['accountId'] not in synchronizing_accounts):
                synchronizing_accounts.append(account_data['accountId'])
        return len(synchronizing_accounts) < self._maxConcurrentSynchronizations

    def remove_synchronization_id(self, synchronization_id: str):
        """Removes synchronization id from slots and removes ids for the same account from the queue.

        Args:
            synchronization_id: Synchronization id.
        """
        if synchronization_id in self._accountsBySynchronizationIds:
            account_id = self._accountsBySynchronizationIds[synchronization_id]['accountId']
            instance_index = self._accountsBySynchronizationIds[synchronization_id]['instanceIndex']
            for key in list(self._accountsBySynchronizationIds.keys()):
                if self._accountsBySynchronizationIds[key]['accountId'] == account_id and \
                        self._accountsBySynchronizationIds[key]['instanceIndex'] == instance_index:
                    self._remove_from_queue(key)
                    del self._accountsBySynchronizationIds[key]
        if synchronization_id in self._synchronizationIds:
            del self._synchronizationIds[synchronization_id]
        self._advance_queue()

    def on_disconnect(self):
        """Clears synchronization ids on disconnect."""
        self._synchronizationIds = {}
        self._advance_queue()

    def _advance_queue(self):
        if self.is_synchronization_available and len(self._synchronizationQueue):
            if not self._synchronizationQueue[0]['promise'].done():
                self._synchronizationQueue[0]['promise'].set_result(True)

    def _remove_from_queue(self, synchronization_id: str):
        for i in range(len(self._synchronizationQueue)):
            if self._synchronizationQueue[i]['synchronizationId'] == synchronization_id and \
              not self._synchronizationQueue[i]['promise'].done():
                self._synchronizationQueue[i]['promise'].set_result(False)
        self._synchronizationQueue = deque(filter(lambda item: item['synchronizationId'] != synchronization_id,
                                                  self._synchronizationQueue))

    async def _process_queue_job(self):
        while len(self._synchronizationQueue) and (len(self._synchronizationIds.values()) <
                                                   self._maxConcurrentSynchronizations):
            await self._synchronizationQueue[0]['promise']
            self._synchronizationQueue.popleft()

    async def schedule_synchronize(self, account_id: str, request: Dict):
        """Schedules to send a synchronization request for account.

        Args:
            account_id: Account id.
            request: Request to send.
        """
        synchronization_id = request['requestId']
        for key in list(self._accountsBySynchronizationIds.keys()):
            if self._accountsBySynchronizationIds[key]['accountId'] == account_id and \
                     self._accountsBySynchronizationIds[key]['instanceIndex'] == \
                    (request['instanceIndex'] if 'instanceIndex' in request else None):
                self.remove_synchronization_id(key)
        self._accountsBySynchronizationIds[synchronization_id] = {
            'accountId': account_id, 'instanceIndex': request['instanceIndex'] if 'instanceIndex' in request else None
        }
        if not self.is_synchronization_available:
            request_resolve = asyncio.Future()
            self._synchronizationQueue.append({
                'synchronizationId': synchronization_id,
                'promise': request_resolve,
                'queueTime': datetime.now().timestamp()
            })
            result = await request_resolve
            if not result:
                return None
        self.update_synchronization_id(synchronization_id)
        return await self._client._rpc_request(account_id, request)



from .connectionRegistry import ConnectionRegistry
from .memoryHistoryStorageModel import MemoryHistoryStorageModel
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from ..clients.metaApi.reconnectListener import ReconnectListener
from ..metaApi.models import MetatraderOrder
from ..metaApi.metatraderAccount import MetatraderAccount
from .metaApiConnection import MetaApiConnection
from .models import MetatraderDeal
from mock import MagicMock, AsyncMock, patch
from datetime import datetime
import pytest


class MockClient(MetaApiWebsocketClient):
    async def subscribe(self, account_id: str, instance_index: int = None):
        pass

    def add_synchronization_listener(self, account_id: str, listener):
        pass

    def add_reconnect_listener(self, listener: ReconnectListener):
        pass


class MockStorage(MemoryHistoryStorageModel):

    def __init__(self):
        super().__init__()
        self._deals = []
        self._historyOrders = []

    @property
    def deals(self):
        return self._deals

    @property
    def history_orders(self):
        return self._historyOrders

    @property
    def last_deal_time_by_instance_index(self):
        return {}

    @property
    def last_history_order_time_by_instance_index(self):
        return {}

    async def clear(self):
        pass

    def last_deal_time(self, instance_index: int = None) -> datetime:
        pass

    def last_history_order_time(self, instance_index: int = None) -> datetime:
        pass

    def on_deal_added(self, instance_index: int, deal: MetatraderDeal):
        pass

    async def load_data_from_disk(self):
        return {'deals': [], 'history_orders': []}

    def on_history_order_added(self, instance_index: int, history_order: MetatraderOrder):
        pass


mock_client: MockClient = None
mock_storage: MockStorage = None
registry: ConnectionRegistry = None


class MockAccount(MetatraderAccount):

    def __init__(self, id):
        super().__init__(MagicMock(), MagicMock(), MagicMock(), MagicMock())
        self._id = id

    @property
    def id(self):
        return self._id


def create_connection_mock():
    mock = MagicMock()
    mock.initialize = AsyncMock()
    mock.subscribe = AsyncMock()
    return mock


@pytest.fixture(autouse=True)
async def run_around_tests():
    global mock_client
    mock_client = MockClient('token')
    global mock_storage
    mock_storage = MockStorage()
    global registry
    registry = ConnectionRegistry(mock_client)
    yield


class TestConnectionRegistry:

    @pytest.mark.asyncio
    async def test_connect_and_add(self):
        """Should connect and add connection to registry."""
        with patch('lib.metaApi.connectionRegistry.MetaApiConnection') as mock_connection:
            connection_instance = create_connection_mock()
            mock_connection.return_value = connection_instance
            account = MockAccount('id')
            connection = await registry.connect(account, mock_storage)
            assert connection.history_storage == connection_instance.history_storage
            connection_instance.initialize.assert_called()
            connection_instance.subscribe.assert_called()
            assert 'id' in registry._connections
            assert registry._connections['id'] == connection

    @pytest.mark.asyncio
    async def test_connect_and_return_previous(self):
        """Should return the same connection on second connect if same account id."""
        with patch('lib.metaApi.connectionRegistry.MetaApiConnection') as mock_connection:
            connection_mock1 = create_connection_mock()
            connection_mock2 = create_connection_mock()
            connection_mock3 = create_connection_mock()
            mock_connection.side_effect = [connection_mock1, connection_mock2, connection_mock3]
            accounts = [MockAccount('id0'), MockAccount('id1')]
            connection0 = await registry.connect(accounts[0], mock_storage)
            connection02 = await registry.connect(accounts[0], mock_storage)
            connection1 = await registry.connect(accounts[1], mock_storage)
            connection_mock1.initialize.assert_called()
            connection_mock1.subscribe.assert_called()
            connection_mock2.initialize.assert_called()
            connection_mock2.subscribe.assert_called()
            assert registry._connections['id0'] == connection0
            assert registry._connections['id1'] == connection1
            assert connection0 == connection02
            assert connection0 != connection1

    @pytest.mark.asyncio
    async def test_remove(self):
        """Should remove the account from registry."""
        with patch('lib.metaApi.connectionRegistry.MetaApiConnection') as mock_connection:
            connection_instance = create_connection_mock()
            mock_connection.return_value = connection_instance
            accounts = [MockAccount('id0'), MockAccount('id1')]
            connection0 = await registry.connect(accounts[0], mock_storage)
            connection1 = await registry.connect(accounts[1], mock_storage)
            assert registry._connections['id0'] == connection0
            assert registry._connections['id1'] == connection1
            registry.remove(accounts[0].id)
            assert not accounts[0].id in registry._connections



class TradeException(Exception):
    """Exception which indicates that a trade has failed."""

    def __init__(self, message: str, numeric_code: int, string_code: str):
        """Inits the timeout exception

        Args:
            message: Exception message.
            numeric_code: Numeric error code.
            string_code: String error code.
        """
        super().__init__(message)
        self.numericCode = numeric_code
        self.stringCode = string_code



from .methodAccessException import MethodAccessException


class MetaApiClient:
    """metaapi.cloud MetaTrader API client."""

    def __init__(self, http_client, token: str, domain: str = 'agiliumtrade.agiliumtrade.ai'):
        """Inits MetaTrader API client instance.

        Args:
            http_client: HTTP client.
            token: Authorization token.
            domain: Domain to connect to, default is agiliumtrade.agiliumtrade.ai.
        """
        self._httpClient = http_client
        self._host = f'https://mt-provisioning-api-v1.{domain}'
        self._token = token

    @property
    def _token_type(self) -> str:
        """Returns type of current token.

        Returns:
            Type of current token.
        """
        if isinstance(self._token, str) and len(self._token.split('.')) == 3:
            return 'api'
        if isinstance(self._token, str) and len(self._token.split('.')) == 1:
            return 'account'
        return ''

    def _is_not_jwt_token(self) -> bool:
        """Checks that current token is not api token.

        Returns:
            Indicator of absence api token.
        """
        return (not isinstance(self._token, str)) or len(self._token.split('.')) != 3

    def _is_not_account_token(self) -> bool:
        """Checks that current token is not account token.

        Returns:
            Indicator of absence account token.
        """
        return (not isinstance(self._token, str)) or len(self._token.split('.')) != 1

    def _handle_no_access_exception(self, method_name):
        """Handles no accessing to the method.

        Args:
            method_name: Name of the method.
        """
        raise MethodAccessException(method_name, self._token_type)



from typing_extensions import TypedDict
from datetime import datetime


class ResponseTimestamps(TypedDict):
    """Object containing request latency information."""
    clientProcessingStarted: datetime
    """Time when request processing has started on client side."""
    serverProcessingStarted: datetime
    """Time when request processing has started on server side."""
    serverProcessingFinished: datetime
    """Time when request processing has finished on server side."""
    clientProcessingFinished: datetime
    """Time when request processing has finished on client side."""


class SymbolPriceTimestamps(TypedDict):
    """Object containing request latency information."""
    eventGenerated: datetime
    """Time the event was generated on exchange side."""
    serverProcessingStarted: datetime
    """Time when request processing has started on server side."""
    serverProcessingFinished: datetime
    """Time when request processing has finished on server side."""
    clientProcessingFinished: datetime
    """Time when request processing has finished on client side."""


class UpdateTimestamps(TypedDict):
    """Timestamps object containing latency information about update streaming."""
    eventGenerated: datetime
    """Time the event was generated on exchange side."""
    serverProcessingStarted: datetime
    """Time when request processing has started on server side."""
    serverProcessingFinished: datetime
    """Time when request processing has finished on server side."""
    clientProcessingFinished: datetime
    """Time when request processing has finished on client side."""


class TradeTimestamps(TypedDict):
    """Timestamps object containing latency information about update streaming."""
    clientProcessingStarted: datetime
    """Time when request processing has started on client side."""
    serverProcessingStarted: datetime
    """Time when request processing has started on server side."""
    serverProcessingFinished: datetime
    """Time when request processing has finished on server side."""
    clientProcessingFinished: datetime
    """Time when request processing has finished on client side."""
    tradeStarted: datetime
    """Time the trade execution was started on server side."""
    tradeExecuted: datetime
    """Time the trade was executed on exchange side."""


class LatencyListener:
    async def on_response(self, account_id: str, type: str, timestamps: ResponseTimestamps):
        """Invoked with latency information when application receives a response to RPC request.

        Args:
            account_id: Account id.
            type: Request type.
            timestamps: Request timestamps object containing latency information.

        Returns:
            A coroutine which resolves when latency information is processed."""
        pass

    async def on_symbol_price(self, account_id: str, symbol: str, timestamps: SymbolPriceTimestamps):
        """Invoked with latency information when application receives symbol price update event.

        Args:
            account_id: Account id.
            symbol: Price symbol.
            timestamps: Timestamps object containing latency information about price streaming.

        Returns:
            A coroutine which resolves when latency information is processed.
        """
        pass

    async def on_update(self, account_id: str, timestamps: UpdateTimestamps):
        """Invoked with latency information when application receives update event.

        Args:
            account_id: Account id.
            timestamps: Timestamps object containing latency information about update streaming.

        Returns:
            A coroutine which resolves when latency information is processed."""
        pass

    async def on_trade(self, account_id: str, timestamps: TradeTimestamps):
        """Invoked with latency information when application receives trade response.

        Args:
            account_id: Account id.
            timestamps: Timestamps object containing latency information about a trade.

        Returns:
            A coroutine which resolves when latency information is processed."""
        pass



from .metatraderAccount import MetatraderAccount
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from ..clients.metaApi.metatraderAccount_client import MetatraderAccountClient, NewMetatraderAccountDto, AccountsFilter
from .connectionRegistryModel import ConnectionRegistryModel
from typing import List


class MetatraderAccountApi:
    """Exposes MetaTrader account API logic to the consumers."""

    def __init__(self, metatrader_account_client: MetatraderAccountClient,
                 meta_api_websocket_client: MetaApiWebsocketClient,
                 connection_registry: ConnectionRegistryModel):
        """Inits a MetaTrader account API instance.

        Args:
            metatrader_account_client: MetaTrader account REST API client.
            meta_api_websocket_client: MetaApi websocket client.
            connection_registry: MetaTrader account connection registry.
        """
        self._metatraderAccountClient = metatrader_account_client
        self._metaApiWebsocketClient = meta_api_websocket_client
        self._connectionRegistry = connection_registry

    async def get_accounts(self, accounts_filter: AccountsFilter = None) -> List[MetatraderAccount]:
        """Retrieves MetaTrader accounts.

        Args:
            accounts_filter: Optional filter.

        Returns:
            A coroutine resolving with an array of MetaTrader account entities.
        """
        accounts = await self._metatraderAccountClient.get_accounts(accounts_filter or {})
        if 'items' in accounts:
            accounts = accounts['items']
        return list(map(lambda account: MetatraderAccount(account, self._metatraderAccountClient,
                                                          self._metaApiWebsocketClient, self._connectionRegistry),
                        accounts))

    async def get_account(self, account_id) -> MetatraderAccount:
        """Retrieves a MetaTrader account by id.

        Args:
            account_id: MetaTrader account id.

        Returns:
            A coroutine resolving with MetaTrader account entity.
        """
        account = await self._metatraderAccountClient.get_account(account_id)
        return MetatraderAccount(account, self._metatraderAccountClient, self._metaApiWebsocketClient,
                                 self._connectionRegistry)

    async def get_account_by_token(self) -> MetatraderAccount:
        """Retrieves a MetaTrader account by token.

        Returns:
            A coroutine resolving with MetaTrader account entity.
        """
        account = await self._metatraderAccountClient.get_account_by_token()
        return MetatraderAccount(account, self._metatraderAccountClient, self._metaApiWebsocketClient,
                                 self._connectionRegistry)

    async def create_account(self, account: NewMetatraderAccountDto) -> MetatraderAccount:
        """Creates a MetaTrader account.

        Args:
            account: MetaTrader account data.

        Returns:
            A coroutine resolving with MetaTrader account entity.
        """
        id = await self._metatraderAccountClient.create_account(account)
        return await self.get_account(id['id'])



from ..metaApi_client import MetaApiClient
from ...metaApi.models import random_id, date
from .copyFactory_models import StrategyId, CopyFactoryAccountUpdate, CopyFactoryStrategyUpdate, \
    ResynchronizationTask, CopyFactoryAccount, CopyFactoryStrategy, CopyFactoryPortfolioStrategy, \
    CopyFactoryPortfolioStrategyUpdate
from ..timeoutException import TimeoutException
from datetime import datetime
import asyncio
from typing import List


class ConfigurationClient(MetaApiClient):
    """metaapi.cloud CopyFactory configuration API (trade copying configuration API) client (see
    https://trading-api-v1.project-stock.agiliumlabs.cloud/swagger/#/)"""

    def __init__(self, http_client, token: str, domain: str = 'agiliumtrade.agiliumtrade.ai'):
        """Inits CopyFactory configuration API client instance.

        Args:
            http_client: HTTP client.
            token: Authorization token.
            domain: Domain to connect to, default is agiliumtrade.agiliumtrade.ai.
        """
        super().__init__(http_client, token, domain)
        self._host = f'https://trading-api-v1.{domain}'

    async def generate_strategy_id(self) -> StrategyId:
        """Retrieves new unused strategy id. Method is accessible only with API access token. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/get_users_current_configuration_unused_strategy_id

        Returns:
            A coroutine resolving with strategy id generated.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('generate_strategy_id')
        opts = {
            'url': f"{self._host}/users/current/configuration/unused-strategy-id",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    def generate_account_id(self) -> str:
        """Generates random account id.

        Returns:
            Account id.
        """
        return random_id(64)

    async def get_accounts(self) -> 'List[CopyFactoryAccount]':
        """Retrieves CopyFactory copy trading accounts. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/get_users_current_configuration_accounts

        Returns:
            A coroutine resolving with CopyFactory accounts found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_accounts')
        opts = {
            'url': f"{self._host}/users/current/configuration/accounts",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def update_account(self, id: str, account: CopyFactoryAccountUpdate):
        """Updates a CopyFactory trade copying account. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /put_users_current_configuration_accounts_accountId

        Args:
            id: Copy trading account id.
            account: Trading account update.

        Returns:
            A coroutine resolving when account is updated.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('update_account')
        opts = {
            'url': f"{self._host}/users/current/configuration/accounts/{id}",
            'method': 'PUT',
            'headers': {
                'auth-token': self._token
            },
            'body': account
        }
        return await self._httpClient.request(opts)

    async def remove_account(self, id: str):
        """Removes a CopyFactory trade copying account. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /delete_users_current_configuration_accounts_accountId

        Args:
            id: Copy trading account id.

        Returns:
            A coroutine resolving when account is removed.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('remove_account')
        opts = {
            'url': f"{self._host}/users/current/configuration/accounts/{id}",
            'method': 'DELETE',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_strategies(self) -> 'List[CopyFactoryStrategy]':
        """Retrieves CopyFactory copy trading strategies. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /get_users_current_configuration_strategies

        Returns:
            A coroutine resolving with CopyFactory strategies found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_strategies')
        opts = {
            'url': f"{self._host}/users/current/configuration/strategies",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def update_strategy(self, id: str, strategy: CopyFactoryStrategyUpdate):
        """Updates a CopyFactory strategy. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /put_users_current_configuration_strategies_strategyId

        Args:
            id: Copy trading strategy id.
            strategy: Trading strategy update.

        Returns:
            A coroutine resolving when strategy is updated.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('update_strategy')
        opts = {
            'url': f"{self._host}/users/current/configuration/strategies/{id}",
            'method': 'PUT',
            'headers': {
                'auth-token': self._token
            },
            'body': strategy
        }
        return await self._httpClient.request(opts)

    async def remove_strategy(self, id: str):
        """Removes a CopyFactory strategy. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default
        /delete_users_current_configuration_strategies_strategyId

        Args:
            id: Copy trading strategy id.

        Returns:
            A coroutine resolving when strategy is removed.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('remove_strategy')
        opts = {
            'url': f"{self._host}/users/current/configuration/strategies/{id}",
            'method': 'DELETE',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_portfolio_strategies(self) -> 'List[CopyFactoryPortfolioStrategy]':
        """Retrieves CopyFactory copy portfolio strategies. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/
        get_users_current_configuration_portfolio_strategies

        Returns:
            A coroutine resolving with CopyFactory portfolio strategies found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_portfolio_strategies')
        opts = {
            'url': f"{self._host}/users/current/configuration/portfolio-strategies",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def update_portfolio_strategy(self, id: str, strategy: CopyFactoryPortfolioStrategyUpdate):
        """Updates a CopyFactory portfolio strategy. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/
        put_users_current_configuration_portfolio_strategies_portfolioId

        Args:
            id: Copy trading portfolio strategy id.
            strategy: Portfolio strategy update.

        Returns:
            A coroutine resolving when portfolio strategy is updated.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('update_portfolio_strategy')
        opts = {
            'url': f"{self._host}/users/current/configuration/portfolio-strategies/{id}",
            'method': 'PUT',
            'headers': {
                'auth-token': self._token
            },
            'body': strategy
        }
        return await self._httpClient.request(opts)

    async def remove_portfolio_strategy(self, id: str):
        """Deletes a CopyFactory portfolio strategy. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/
        delete_users_current_configuration_portfolio_strategies_portfolioId

        Args:
            id: Portfolio strategy id.

        Returns:
            A coroutine resolving when portfolio strategy is removed.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('remove_portfolio_strategy')
        opts = {
            'url': f"{self._host}/users/current/configuration/portfolio-strategies/{id}",
            'method': 'DELETE',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_active_resynchronization_tasks(self, connection_id) -> 'List[ResynchronizationTask]':
        """Returns list of active resynchronization tasks for a specified connection. See
        https://trading-api-v1.agiliumtrade.agiliumtrade.ai/swagger/#!/default/
        get_users_current_configuration_connections_connectionId_active_resynchronization_tasks

        Args:
            connection_id: MetaApi account id to return tasks for.

        Returns:
            A coroutine resolving with list of active resynchronization tasks.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_active_resynchronization_tasks')
        opts = {
            'url': f"{self._host}/users/current/configuration/connections/{connection_id}" +
                   "/active-resynchronization-tasks",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }

        tasks = await self._httpClient.request(opts)
        for task in tasks:
            task['createdAt'] = date(task['createdAt'])
        return tasks

    async def wait_resynchronization_tasks_completed(self, connection_id: str, timeout_in_seconds: float = 300,
                                                     interval_in_milliseconds: float = 1000):
        """Waits until active resynchronization tasks are completed.

        Args:
            connection_id: MetaApi account id to wait tasks completed for.
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between tasks reload while waiting for a change, default is 1s.

        Returns:
            A couroutine which resolves when tasks are completed.

        Raises:
            TimeoutException: If tasks have not completed within timeout allowed.
        """
        start_time = datetime.now().timestamp()
        tasks = await self.get_active_resynchronization_tasks(connection_id)
        while len(tasks) and (start_time + timeout_in_seconds) > datetime.now().timestamp():
            await asyncio.sleep(interval_in_milliseconds / 1000)
            tasks = await self.get_active_resynchronization_tasks(connection_id)
        if len(tasks):
            raise TimeoutException('Timed out waiting for resynchronization tasks for account '
                                   + connection_id + ' to be completed')



from ...metaApi.models import MetatraderPosition, MetatraderAccountInformation, MetatraderOrder, \
    MetatraderDeal, MetatraderSymbolSpecification, MetatraderSymbolPrice
from abc import ABC
from typing import List, Optional
from typing_extensions import TypedDict


class HealthStatus(TypedDict):
    """Server-side application health status."""
    restApiHealthy: Optional[bool]
    """Flag indicating that REST API is healthy."""
    copyFactorySubscriberHealthy: Optional[bool]
    """Flag indicating that CopyFactory subscriber is healthy."""
    copyFactoryProviderHealthy: Optional[bool]
    """Flag indicating that CopyFactory provider is healthy."""


class SynchronizationListener(ABC):
    """Defines interface for a synchronization listener class."""

    async def on_connected(self, instance_index: int, replicas: int):
        """Invoked when connection to MetaTrader terminal established.

        Args:
            instance_index: Index of an account instance connected.
            replicas: Number of account replicas launched.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_health_status(self, instance_index: int, status: HealthStatus):
        """Invoked when a server-side application health status is received from MetaApi.

        Args:
            instance_index: Index of an account instance connected.
            status: Server-side application health status.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_disconnected(self, instance_index: int):
        """Invoked when connection to MetaTrader terminal terminated.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
             A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_broker_connection_status_changed(self, instance_index: int, connected: bool):
        """Invoked when broker connection status have changed.

        Args:
            instance_index: Index of an account instance connected.
            connected: Is MetaTrader terminal is connected to broker.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_synchronization_started(self, instance_index: int):
        """Invoked when MetaTrader terminal state synchronization is started.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_account_information_updated(self, instance_index: int,
                                             account_information: MetatraderAccountInformation):
        """Invoked when MetaTrader position is updated.

        Args:
            instance_index: Index of an account instance connected.
            account_information: Updated MetaTrader position.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_positions_replaced(self, instance_index: int, positions: List[MetatraderPosition]):
        """Invoked when the positions are replaced as a result of initial terminal state synchronization.

        Args:
            instance_index: Index of an account instance connected.
            positions: Updated array of positions.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_position_updated(self, instance_index: int, position: MetatraderPosition):
        """Invoked when MetaTrader position is updated.

        Args:
            instance_index: Index of an account instance connected.
            position: Updated MetaTrader position.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_position_removed(self, instance_index: int, position_id: str):
        """Invoked when MetaTrader position is removed.

        Args:
            instance_index: Index of an account instance connected.
            position_id: Removed MetaTrader position id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_orders_replaced(self, instance_index: int, orders: List[MetatraderOrder]):
        """Invoked when the orders are replaced as a result of initial terminal state synchronization.

        Args:
            instance_index: Index of an account instance connected.
            orders: Updated array of orders.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_order_updated(self, instance_index: int, order: MetatraderOrder):
        """Invoked when MetaTrader order is updated.

        Args:
            instance_index: Index of an account instance connected.
            order: Updated MetaTrader order.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_order_completed(self, instance_index: int, order_id: str):
        """Invoked when MetaTrader order is completed (executed or canceled).

        Args:
            instance_index: Index of an account instance connected.
            order_id: Completed MetaTrader order id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_history_order_added(self, instance_index: int, history_order: MetatraderOrder):
        """Invoked when a new MetaTrader history order is added.

        Args:
            instance_index: Index of an account instance connected.
            history_order: New MetaTrader history order.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_deal_added(self, instance_index: int, deal: MetatraderDeal):
        """Invoked when a new MetaTrader history deal is added.

        Args:
            instance_index: Index of an account instance connected.
            deal: New MetaTrader history deal.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_deal_synchronization_finished(self, instance_index: int, synchronization_id: str):
        """Invoked when a synchronization of history deals on a MetaTrader account have finished.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Synchronization request id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_order_synchronization_finished(self, instance_index: int, synchronization_id: str):
        """Invoked when a synchronization of history orders on a MetaTrader account have finished.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Synchronization request id.

        Returns:
             A coroutine which resolves when the asynchronous event is processed
        """
        pass

    async def on_symbol_specification_updated(self, instance_index: int, specification: MetatraderSymbolSpecification):
        """Invoked when a symbol specification was updated

        Args:
            instance_index: Index of an account instance connected.
            specification: Updated MetaTrader symbol specification.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_symbol_price_updated(self, instance_index: int, price: MetatraderSymbolPrice):
        """Invoked when a symbol price was updated.

        Args:
            instance_index: Index of an account instance connected.
            price: Updated MetaTrader symbol price.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass

    async def on_symbol_prices_updated(self, instance_index: int, prices: List[MetatraderSymbolPrice],
                                       equity: float = None, margin: float = None, free_margin: float = None,
                                       margin_level: float = None):
        """Invoked when prices for several symbols were updated.

        Args:
            instance_index: Index of an account instance connected.
            prices: Updated MetaTrader symbol prices.
            equity: Account liquidation value.
            margin: Margin used.
            free_margin: Free margin.
            margin_level: Margin level calculated as % of equity/margin.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        pass



from ..clients.metaApi.metatraderAccount_client import MetatraderAccountClient, MetatraderAccountDto, \
    MetatraderAccountUpdateDto, Extension
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from ..clients.timeoutException import TimeoutException
from .metaApiConnection import MetaApiConnection
from .metatraderAccountModel import MetatraderAccountModel
from ..metaApi.historyFileManager import HistoryFileManager
from .historyStorage import HistoryStorage
from .connectionRegistryModel import ConnectionRegistryModel
from datetime import datetime, timedelta
from typing import List, Dict
import asyncio


class MetatraderAccount(MetatraderAccountModel):
    """Implements a MetaTrader account entity"""

    def __init__(self, data: MetatraderAccountDto, metatrader_account_client: MetatraderAccountClient,
                 meta_api_websocket_client: MetaApiWebsocketClient, connection_registry: ConnectionRegistryModel):
        """Inits a MetaTrader account entity.

        Args:
            data: MetaTrader account data.
            metatrader_account_client: MetaTrader account REST API client.
            meta_api_websocket_client: MetaApi websocket client.
            connection_registry: Metatrader account connection registry.
        """
        self._data = data
        self._metatraderAccountClient = metatrader_account_client
        self._metaApiWebsocketClient = meta_api_websocket_client
        self._connectionRegistry = connection_registry

    @property
    def id(self) -> str:
        """Returns account id.

        Returns:
            Account id.
        """
        return self._data['_id']

    @property
    def name(self) -> str:
        """Returns account name.

        Returns:
            Account name.
        """
        return self._data['name']

    @property
    def type(self) -> str:
        """Returns account type. Possible values are cloud and self-hosted.

        Returns:
            Account type.
        """
        return self._data['type']

    @property
    def login(self) -> str:
        """Returns account login.

        Returns:
            Account login.
        """
        return self._data['login']

    @property
    def server(self) -> str:
        """Returns MetaTrader server which hosts the account.

        Returns:
            MetaTrader server which hosts the account.
        """
        return self._data['server']

    @property
    def provisioning_profile_id(self) -> str:
        """Returns id of the account's provisioning profile.

        Returns:
            Id of the account's provisioning profile.
        """
        return self._data['provisioningProfileId']

    @property
    def application(self) -> str:
        """Returns application name to connect the account to. Currently allowed values are MetaApi and AgiliumTrade.

        Returns:
            Application name to connect the account to.
        """
        return self._data['application']

    @property
    def magic(self) -> int:
        """Returns MetaTrader magic to place trades using.

        Returns:
            MetaTrader magic to place trades using.
        """
        return self._data['magic']

    @property
    def state(self) -> str:
        """Returns account deployment state. One of CREATED, DEPLOYING, DEPLOYED, UNDEPLOYING, UNDEPLOYED, DELETING

        Returns:
            Account deployment state.
        """
        return self._data['state']

    @property
    def connection_status(self) -> str:
        """Returns terminal & broker connection status, one of CONNECTED, DISCONNECTED, DISCONNECTED_FROM_BROKER

        Returns:
            Terminal & broker connection status.
        """
        return self._data['connectionStatus']

    @property
    def access_token(self) -> str:
        """Returns authorization access token to be used for accessing single account data.
        Intended to be used in browser API.

        Returns:
            Authorization token.
        """
        return self._data['accessToken']

    @property
    def manual_trades(self) -> bool:
        """Returns flag indicating if trades should be placed as manual trades on this account.

        Returns:
            Flag indicating if trades should be placed as manual trades on this account.
        """
        return 'manualTrades' in self._data and self._data['manualTrades']

    @property
    def extensions(self) -> List[Extension]:
        """Returns API extensions.

        Returns:
            API extensions.
        """
        return self._data['extensions']

    @property
    def metadata(self) -> Dict:
        """Returns extra information which can be stored together with your account.

        Returns:
            Extra information which can be stored together with your account.
        """
        return self._data['metadata']

    @property
    def reliability(self) -> str:
        """Returns reliability value. Possible values are regular and high.

        Returns:
            Reliability value.
        """
        return self._data['reliability']

    async def reload(self):
        """Reloads MetaTrader account from API.

        Returns:
            A coroutine resolving when MetaTrader account is updated.
        """
        self._data = await self._metatraderAccountClient.get_account(self.id)

    async def remove(self):
        """Removes MetaTrader account. Cloud account transitions to DELETING state.
        It takes some time for an account to be eventually deleted. Self-hosted account is deleted immediately.

        Returns:
            A coroutine resolving when account is scheduled for deletion.
        """
        self._connectionRegistry.remove(self.id)
        await self._metatraderAccountClient.delete_account(self.id)
        file_manager = HistoryFileManager(self.id, self.application)
        await file_manager.delete_storage_from_disk()
        if self.type != 'self-hosted':
            try:
                await self.reload()
            except Exception as err:
                if err.__class__.__name__ != 'NotFoundException':
                    raise err

    async def deploy(self):
        """Schedules account for deployment. It takes some time for API server to be started and account to reach the
        DEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for deployment.
        """
        await self._metatraderAccountClient.deploy_account(self.id)
        await self.reload()

    async def undeploy(self):
        """Schedules account for undeployment. It takes some time for API server to be stopped and account to reach the
        UNDEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for undeployment.
        """
        self._connectionRegistry.remove(self.id)
        await self._metatraderAccountClient.undeploy_account(self.id)
        await self.reload()

    async def redeploy(self):
        """Schedules account for redeployment. It takes some time for API server to be restarted and account to reach
        the DEPLOYED state.

        Returns:
            A coroutine resolving when account is scheduled for redeployment.
        """
        await self._metatraderAccountClient.redeploy_account(self.id)
        await self.reload()

    async def wait_deployed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has finished deployment and account reached the DEPLOYED state.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is deployed.

        Raises:
            TimeoutException: If account has not reached the DEPLOYED state within timeout allowed.
        """
        start_time = datetime.now()
        await self.reload()
        while self.state != 'DEPLOYED' and (start_time + timedelta(seconds=timeout_in_seconds) > datetime.now()):
            await self._delay(interval_in_milliseconds)
            await self.reload()
        if self.state != 'DEPLOYED':
            raise TimeoutException('Timed out waiting for account ' + self.id + ' to be deployed')

    async def wait_undeployed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has finished undeployment and account reached the UNDEPLOYED state.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is undeployed.

        Raises:
            TimeoutException: If account have not reached the UNDEPLOYED state within timeout allowed.
        """
        start_time = datetime.now()
        await self.reload()
        while self.state != 'UNDEPLOYED' and (start_time + timedelta(seconds=timeout_in_seconds) > datetime.now()):
            await self._delay(interval_in_milliseconds)
            await self.reload()
        if self.state != 'UNDEPLOYED':
            raise TimeoutException('Timed out waiting for account ' + self.id + ' to be undeployed')

    async def wait_removed(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until account has been deleted.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m.
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when account is deleted.

        Raises:
            TimeoutException: If account was not deleted within timeout allowed.
        """
        start_time = datetime.now()
        try:
            await self.reload()
            while (start_time + timedelta(seconds=timeout_in_seconds)) > datetime.now():
                await self._delay(interval_in_milliseconds)
                await self.reload()
            raise TimeoutException('Timed out waiting for account ' + self.id + ' to be deleted')
        except Exception as err:
            if err.__class__.__name__ == 'NotFoundException':
                return
            else:
                raise err

    async def wait_connected(self, timeout_in_seconds=300, interval_in_milliseconds=1000):
        """Waits until API server has connected to the terminal and terminal has connected to the broker.

        Args:
            timeout_in_seconds: Wait timeout in seconds, default is 5m
            interval_in_milliseconds: Interval between account reloads while waiting for a change, default is 1s.

        Returns:
            A coroutine which resolves when API server is connected to the broker.

        Raises:
            TimeoutException: If account has not connected to the broker within timeout allowed.
        """
        start_time = datetime.now()
        await self.reload()
        while self.connection_status != 'CONNECTED' and (start_time +
                                                         timedelta(seconds=timeout_in_seconds)) > datetime.now():
            await self._delay(interval_in_milliseconds)
            await self.reload()
        if self.connection_status != 'CONNECTED':
            raise TimeoutException('Timed out waiting for account ' + self.id + ' to connect to the broker')

    async def connect(self, history_storage: HistoryStorage = None, history_start_time: datetime = None) -> \
            MetaApiConnection:
        """Connects to MetaApi.

        Args:
            history_storage: Optional history storage.
            history_start_time: History start time. Used for tests.

        Returns:
            MetaApi connection.
        """
        return await self._connectionRegistry.connect(self, history_storage, history_start_time)

    async def update(self, account: MetatraderAccountUpdateDto):
        """Updates MetaTrader account data.

        Args:
            account: MetaTrader account update.

        Returns:
            A coroutine resolving when account is updated.
        """
        await self._metatraderAccountClient.update_account(self.id, account)
        await self.reload()

    async def _delay(self, timeout_in_milliseconds):
        await asyncio.sleep(timeout_in_milliseconds / 1000)



import os
import asyncio
from metaapi_cloud_sdk import MetaApi, CopyFactory

# your MetaApi API token
token = os.getenv('TOKEN') or '<put in your token here>'
# your master MetaApi account id
master_account_id = os.getenv('MASTER_ACCOUNT_ID') or '<put in your account id here>'
# your slave MetaApi account id
slave_account_id = os.getenv('SLAVE_ACCOUNT_ID') or '<put in your account id here>'


async def configure_copyfactory():
    api = MetaApi(token)
    copy_factory = CopyFactory(token)

    try:
        global master_account_id
        global slave_account_id
        accounts = await api.metatrader_account_api.get_accounts()
        master_metaapi_account = next((a for a in accounts if a.id == master_account_id), None)
        if (not hasattr(master_metaapi_account, 'application')) or master_metaapi_account.application != 'CopyFactory':
            raise Exception('Please specify CopyFactory application field value in your MetaApi '
                            'account in order to use it in CopyFactory API')

        slave_metaapi_account = next((a for a in accounts if a.id == slave_account_id), None)
        if (not hasattr(slave_metaapi_account, 'application')) or slave_metaapi_account.application != 'CopyFactory':
            raise Exception('Please specify CopyFactory application field value in your MetaApi '
                            'account in order to use it in CopyFactory API')

        configuration_api = copy_factory.configuration_api
        master_account_id = configuration_api.generate_account_id()
        slave_account_id = configuration_api.generate_account_id()
        await configuration_api.update_account(master_account_id, {
            'name': 'Demo master account',
            'connectionId': master_metaapi_account.id,
            'subscriptions': []
        })

        # create a strategy being copied
        strategy_id = await configuration_api.generate_strategy_id()
        await configuration_api.update_strategy(strategy_id['id'], {
            'name': 'Test strategy',
            'description': 'Some useful description about your strategy',
            'connectionId': master_metaapi_account.id
        })

        # subscribe slave CopyFactory accounts to the strategy
        await configuration_api.update_account(slave_account_id, {
            'name': 'Demo slave account',
            'connectionId': slave_metaapi_account.id,
            'subscriptions': [
                {
                    'strategyId': strategy_id['id'],
                    'multiplier': 1
                }
            ]
        })

        print('Please note that it can take some time for CopyFactory to initialize accounts. During this time the '
              'MetaApi accounts may redeploy a couple of times. After initialization finishes, you can copy trades '
              'from your master to slave account.')
    except Exception as err:
        print(api.format_error(err))


asyncio.run(configure_copyfactory())



import os
import asyncio
from metaapi_cloud_sdk import MetaApi
from metaapi_cloud_sdk.clients.metaApi.synchronizationListener import SynchronizationListener
from metaapi_cloud_sdk.metaApi.models import MetatraderSymbolPrice

token = os.getenv('TOKEN') or '<put in your token here>'
account_id = os.getenv('ACCOUNT_ID') or '<put in your account id here>'



class EURUSDListener(SynchronizationListener):
    async def on_symbol_price_updated(self, instance_index: int, price: MetatraderSymbolPrice):
        if price['symbol'] == 'EURUSD':
            print('EURUSD price updated', price)


async def stream_quotes():
    api = MetaApi(token)
    try:
        account = await api.metatrader_account_api.get_account(account_id)

        #  wait until account is deployed and connected to broker
        print('Deploying account')
        #await account.deploy()
        print('Waiting for API server to connect to broker (may take couple of minutes)')
        await account.wait_connected()

        # connect to MetaApi API
        connection = await account.connect()

        eur_usd_listener = EURUSDListener()
        connection.add_synchronization_listener(eur_usd_listener)

        # wait until terminal state synchronized to the local state
        print('Waiting for SDK to synchronize to terminal state (may take some time depending on your history size), the price streaming will start once synchronization finishes')
        await connection.wait_synchronized({'timeoutInSeconds': 1200})

        # Add symbol to MarketWatch if not yet added
        await connection.subscribe_to_market_data('EURUSD')

        print('Streaming EURUSD price now...')

        while True:
            await asyncio.sleep(1)

    except Exception as err:
        print(api.format_error(err))

asyncio.run(stream_quotes())



from .packetOrderer import PacketOrderer
import pytest
import asyncio
from mock import MagicMock, patch
from datetime import datetime

out_of_order_listener = MagicMock()
packet_orderer = PacketOrderer(out_of_order_listener, 0.5)
date = datetime.fromtimestamp(1000000000)


@pytest.fixture(autouse=True)
async def run_around_tests():
    global packet_orderer
    packet_orderer = PacketOrderer(out_of_order_listener, 0.5)
    packet_orderer.start()
    yield
    packet_orderer.stop()


class TestPacketOrderer:
    @pytest.mark.asyncio
    async def test_no_sequence_number(self):
        """Should return packets without a sequence number out immediately."""
        packet_without_sn = {
            'type': 'authenticated',
            'connectionId': 'accountId',
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(packet_without_sn) == [packet_without_sn]

    @pytest.mark.asyncio
    async def test_restore(self):
        """Should restore packet order."""
        first_packet = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267178,
            'sequenceNumber': 13,
            'synchronizationId': 'synchronizationId',
            'accountId': 'accountId'
        }
        second_packet = {
            'type': 'prices',
            'sequenceTimestamp': 1603124267180,
            'sequenceNumber': 14,
            'accountId': 'accountId'
        }
        third_packet = {
            'type': 'accountInformation',
            'sequenceTimestamp': 1603124267187,
            'sequenceNumber': 15,
            'accountId': 'accountId'
        }
        fourth_packet = {
            'type': 'positions',
            'sequenceTimestamp': 1603124267188,
            'sequenceNumber': 16,
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(second_packet) == []
        assert packet_orderer.restore_order(first_packet) == [first_packet, second_packet]
        assert packet_orderer.restore_order(fourth_packet) == []
        assert packet_orderer.restore_order(third_packet) == [third_packet, fourth_packet]

    @pytest.mark.asyncio
    async def test_filter_prev_sync_packets_with_specifications(self):
        """Should filter out packets from previous synchronization attempt that includes synchronization start."""
        previous_start = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267178,
            'sequenceNumber': 13,
            'synchronizationId': 'synchronizationId',
            'accountId': 'accountId'
        }
        one_of_previous_packets = {
            'type': 'positions',
            'sequenceTimestamp': 1603124267188,
            'sequenceNumber': 15,
            'accountId': 'accountId'
        }
        this_specifications = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267198,
            'sequenceNumber': 1,
            'synchronizationId': 'synchronizationId',
            'accountId': 'accountId'
        }
        this_second_packet = {
            'type': 'prices',
            'sequenceTimestamp': 1603124268198,
            'sequenceNumber': 2,
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(previous_start) == [previous_start]
        assert packet_orderer.restore_order(one_of_previous_packets) == []
        assert packet_orderer.restore_order(this_second_packet) == []
        assert packet_orderer.restore_order(this_specifications) == [this_specifications, this_second_packet]

    @pytest.mark.asyncio
    async def test_filter_prev_sync_packets_without_specifications(self):
        """Should filter out packets from previous synchronization attempt that does not includes the start."""
        one_of_previous_packets = {
            'type': 'positions',
            'sequenceTimestamp': 1603124267188,
            'sequenceNumber': 15,
            'accountId': 'accountId'
        }
        this_start = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267198,
            'sequenceNumber': 1,
            'synchronizationId': 'synchronizationId',
            'accountId': 'accountId'
        }
        this_second_packet = {
            'type': 'prices',
            'sequenceTimestamp': 1603124268198,
            'sequenceNumber': 2,
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(one_of_previous_packets) == []
        assert packet_orderer.restore_order(this_second_packet) == []
        assert packet_orderer.restore_order(this_start) == [this_start, this_second_packet]

    @pytest.mark.asyncio
    async def test_duplicate(self):
        """Should pass through duplicate packets."""
        specifications_packet = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267198,
            'sequenceNumber': 16,
            'synchronizationId': 'synchronizationId',
            'accountId': 'accountId'
        }
        second_packet = {
            'type': 'prices',
            'sequenceTimestamp': 1603124268198,
            'sequenceNumber': 17,
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(specifications_packet) == [specifications_packet]
        assert packet_orderer.restore_order(second_packet) == [second_packet]
        assert packet_orderer.restore_order(second_packet) == [second_packet]

    @pytest.mark.asyncio
    async def test_return_in_order(self):
        """Should return in-order packets immediately."""
        first_packet = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267178,
            'sequenceNumber': 13,
            'synchronizationId': 'synchronizationId',
            'accountId': 'accountId'
        }
        second_packet = {
            'type': 'prices',
            'sequenceTimestamp': 1603124267180,
            'sequenceNumber': 14,
            'accountId': 'accountId'
        }
        third_packet = {
            'type': 'accountInformation',
            'sequenceTimestamp': 1603124267187,
            'sequenceNumber': 15,
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(first_packet) == [first_packet]
        assert packet_orderer.restore_order(second_packet) == [second_packet]
        assert packet_orderer.restore_order(third_packet) == [third_packet]

    @pytest.mark.asyncio
    async def test_call_out_of_order_once(self):
        """Should call on-out-of-order listener only once per synchronization attempt."""
        out_of_order_listener.on_out_of_order_packet = MagicMock()
        first_packet = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267178,
            'sequenceNumber': 13,
            'synchronizationId': 'synchronizationId',
            'accountId': 'accountId'
        }
        third_packet = {
            'type': 'orders',
            'sequenceTimestamp': 1603124267193,
            'sequenceNumber': 15,
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(first_packet) == [first_packet]
        assert packet_orderer.restore_order(third_packet) == []
        await asyncio.sleep(1)
        out_of_order_listener.on_out_of_order_packet.assert_called_once()
        args_list = out_of_order_listener.on_out_of_order_packet.call_args_list[0].args
        assert args_list[0] == 'accountId'
        assert args_list[1] == 0
        assert args_list[2] == 14
        assert args_list[3] == 15
        assert args_list[4] == third_packet
        await asyncio.sleep(1)
        out_of_order_listener.on_out_of_order_packet.assert_called_once()

    @pytest.mark.asyncio
    async def test_call_out_of_order_if_timeout(self):
        """Should call on-out-of-order listener if the first packet in wait list is timed out."""
        out_of_order_listener.on_out_of_order_packet = MagicMock()
        timed_out_packet = {
            'accountId': 'accountId',
            'instanceId': 'accountId:0',
            'instanceIndex': 0,
            'sequenceNumber': 11,
            'packet': {},
            'receivedAt': date
        }
        not_timed_out_packet = {
            'accountId': 'accountId',
            'instanceId': 'accountId:0',
            'instanceIndex': 0,
            'sequenceNumber': 15,
            'packet': {},
            'receivedAt': datetime.fromtimestamp(10000000000)
        }
        packet_orderer._sequenceNumberByInstance['accountId:0'] = 1
        packet_orderer._packetsByInstance['accountId:0'] = [
            timed_out_packet,
            not_timed_out_packet
        ]
        await asyncio.sleep(1)
        out_of_order_listener.on_out_of_order_packet.assert_called_once()
        args_list = out_of_order_listener.on_out_of_order_packet.call_args_list[0].args
        assert args_list[0] == 'accountId'
        assert args_list[1] == 0
        assert args_list[2] == 2
        assert args_list[3] == 11
        assert args_list[4] == timed_out_packet['packet']
        await asyncio.sleep(1)
        out_of_order_listener.on_out_of_order_packet.assert_called_once()

    @pytest.mark.asyncio
    async def test_not_call_out_of_order_if_not_timeout(self):
        """Should not call on-out-of-order listener if the first packet in wait list is not timed out."""
        out_of_order_listener.on_out_of_order_packet = MagicMock()
        timed_out_packet = {
            'accountId': 'accountId',
            'sequenceNumber': 11,
            'packet': {},
            'receivedAt': date
        }
        not_timed_out_packet = {
            'accountId': 'accountId',
            'sequenceNumber': 15,
            'packet': {},
            'receivedAt': datetime.fromtimestamp(10000000000)
        }
        packet_orderer._sequenceNumberByInstance['accountId:0'] = 1
        packet_orderer._packetsByInstance['accountId:0'] = [
            not_timed_out_packet,
            timed_out_packet
        ]
        await asyncio.sleep(1)
        out_of_order_listener.on_out_of_order_packet.assert_not_called()

    @pytest.mark.asyncio
    async def test_not_call_out_of_order_if_before_sync_start(self):
        """Should not call on-out-of-order listener for packets that come before synchronization start."""
        out_of_order_listener.on_out_of_order_packet = MagicMock()
        out_of_order_packet = {
            'accountId': 'accountId',
            'sequenceNumber': 11,
            'packet': {},
            'receivedAt': date
        }

        # There were no synchronization start packets
        if 'accountId' in packet_orderer._sequenceNumberByInstance:
            del packet_orderer._sequenceNumberByInstance['accountId:0']

        packet_orderer._packetsByInstance['accountId:0'] = [out_of_order_packet]
        await asyncio.sleep(1)
        out_of_order_listener.on_out_of_order_packet.assert_not_called()

    @pytest.mark.asyncio
    async def test_maintain_fixed_queue(self):
        """Should maintain a fixed queue of wait list."""
        packet_orderer._waitListSizeLimit = 1
        second_packet = {
            'type': 'prices',
            'sequenceTimestamp': 1603124267180,
            'sequenceNumber': 14,
            'accountId': 'accountId'
        }
        third_packet = {
            'type': 'accountInformation',
            'sequenceTimestamp': 1603124267187,
            'sequenceNumber': 15,
            'accountId': 'accountId'
        }
        packet_orderer.restore_order(second_packet)
        assert len(packet_orderer._packetsByInstance['accountId:0']) == 1
        assert packet_orderer._packetsByInstance['accountId:0'][0]['packet'] == second_packet
        packet_orderer.restore_order(third_packet)
        assert len(packet_orderer._packetsByInstance['accountId:0']) == 1
        assert packet_orderer._packetsByInstance['accountId:0'][0]['packet'] == third_packet

    @pytest.mark.asyncio
    async def test_count_start_packets_with_no_sync_id_as_out_of_order(self):
        """Should count start packets with undefined synchronziationId as out-of-order."""
        start_packet = {
            'type': 'synchronizationStarted',
            'sequenceTimestamp': 1603124267198,
            'sequenceNumber': 16,
            'accountId': 'accountId'
        }
        assert packet_orderer.restore_order(start_packet) == []
        assert len(packet_orderer._packetsByInstance['accountId:0']) == 1
        assert packet_orderer._packetsByInstance['accountId:0'][0]['packet'] == start_packet
